/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle <ingo@fargonauten.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.agai.xml.transport;

import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Node;
import nu.xom.ParentNode;
import org.jaxen.JaxenException;
import org.jaxen.XPath;

/**
 *
 * @author iluetkeb
 */
public abstract class AbstractTransportAdapter<T> implements TransportAdapter<T> {
    private final Class<T> adaptableClass;
    private final String name;
    
    /** Create adapter for the given class with the identifier. */
    public AbstractTransportAdapter(Class<T> clazz, String name) {
        this.adaptableClass = clazz;
        this.name = name;
    }
    /** Create Adapter using the class-name as its name. */
    public AbstractTransportAdapter(Class<T> clazz) {
        this(clazz, clazz.getName());
    }

    /**
     * Default implementation of creating a new document uses the
     * adapter name as the name of the root tag.
     */
    @Override
    public Document adapt(T o) {
        Element root = new Element(name);
        Document doc = new Document(root);
        adapt(root, o);
        return doc;
    }

    /**
     * Navigate path starting from context and then call adapt(target, o)
     *
     * @throws TransportConversionException when the path could not be found
     *  or is not a ParentNode
     */
    @Override
    public ParentNode adapt(ParentNode context, XPath path, T o) throws TransportConversionException {
        try {
            return adapt((ParentNode) path.selectSingleNode(context), o);
        } catch (JaxenException ex) {
            throw new TransportConversionException("Could not traverse path " +
                    path + " from context node " + context, ex);
        } catch(ClassCastException ex) {
            throw new TransportConversionException("Target of path " + path +
                    " is not a ParentNode", ex);
        }
    }

    /** Insert o starting from this ParentNode. */
    public abstract ParentNode adapt(ParentNode context, T o);

    /** Unpack type from the given node. */
    public abstract T unpack(Node data) throws TransportConversionException;

    /**
     * Navigate the given path based on the given context and call
     * unpack(Node) there.
     */
    @Override
    public T unpack(Node context, XPath path) throws TransportConversionException {
        try {
            return unpack((Node) path.selectSingleNode(context));
        } catch (JaxenException ex) {
            throw new TransportConversionException("Could not traverse path " +
                    path + " from context " + context);
        } catch(ClassCastException ex) {
            // this should not be able to happen
            throw new TransportConversionException("Target of path " + path +
                    " is not a Node", ex);
        }
    }

    /** Returns the class given on construction. */
    @Override
    public final Class<T> getAdaptableClass() {
        return adaptableClass;
    }

    /** Returns the name given on construction. */
    @Override
    public final String getName() {
        return name;
    }

    /** Default implementation returns this object. */
    public TransportAdapter<T> setContext(TransportContext context) {
        return this;
    }

    /**
     * If data is an Element, returns it unchanged. If it is a Document,
     * returns the root element. Otherwise, throws.
     *
     * @throws TransportConversionException if data is neither Element nor a Document
     */
    protected Element asElement(Node data) throws TransportConversionException {
        if(data instanceof Element)
            return (Element)data;
        else if(data instanceof Document)
            return ((Document)data).getRootElement();
        else
            throw new TransportConversionException("Context item " + data +
                    " is neither Element nor Document.");

    }

    protected Element asElement(ParentNode data) {
        if(data instanceof Element)
            return (Element)data;
        else if(data instanceof Document)
            return ((Document)data).getRootElement();
        else
            throw new IllegalArgumentException("Data cannot be a ParentNode");

    }
}
