/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle <ingo@fargonauten.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.agai.xml.transport;

import de.unibi.agai.xml.*;
import java.util.logging.Logger;
import nu.xom.Attribute;
import nu.xom.Document;
import nu.xom.Node;
import nu.xom.ParentNode;
import org.jaxen.JaxenException;
import org.jaxen.XPath;

/**
 * A TransportAdapter that uses a template to insert data, in the spirit
 * of XMLTIO.
 *
 * @author iluetkeb
 */
public class TemplateTransportAdapter<T,O> implements TransportAdapter<T> {
    private static final Logger logger = Logger.getLogger(TemplateTransportAdapter.class.getName());
    private final Document template;
    private final XPath start;
    private final TypeConverter<T,O> converter;
    private final ValueSplice<O> splice;
    private final Class<T> adaptableClass;
    private final ValuePolicy<O> transportPolicy;
    private final ValuePolicy<T> dataPolicy;

    public TemplateTransportAdapter(Document template, XPath start, 
            ValueSplice<O> splice, TypeConverter<T,O> converter,
            ValuePolicy<O> transportPolicy, ValuePolicy<T> dataPolicy,
            Class<T> clazz) {
        this.template = template;
        this.start = start;
        this.converter = converter;
        this.splice = splice;
        this.adaptableClass = clazz;
        this.transportPolicy = transportPolicy;
        this.dataPolicy = dataPolicy;
    }
    /**
     * Create a template transport adapter that uses a non-empty policy
     * for the transport value and no policy for the data value.
     */
    public TemplateTransportAdapter(Document template, XPath start,
            ValueSplice<O> splice, TypeConverter<T,O> converter,
            Class<T> clazz) throws JaxenException {
        this(template, start, splice, converter, new NonEmptyPolicy(),
                null, clazz);
        // check that the XPath works
        if(start.selectSingleNode(template) == null)
            throw new IllegalArgumentException("XPath not found in template");
    }

    /**
     * Adapts the given data by making a copy of the template, navigating to
     * the specified XPath and calling assignTo(Node, T).
     * @return new XOPData instance containing the template as the document.
     */
    public Document adapt(T o) {
        Document result = (Document)template.copy();
        try {
            adapt(result, start, o);
        } catch(TransportConversionException ex) {
            throw new RuntimeException("Ooops! Start xpath failed on template!", ex);
        }
        return result;
    }

    /**
     * Adapts the given data by navigating to the default XPath in the
     * XOPData's document and calling assignTo(Node, T).
     * 
     * @return The node the content was inserted in. This may or may
     * not be the original target node. If not, it is a child.
     */
    @Override
    public ParentNode adapt(ParentNode target, T o) {
        return splice.assignTo(target, converter.toTransport(o));
    }
    /**
     * Adapts the given data by navigating to the specified XPath in the
     * XOPData's document and calling assignTo(Node, T).
     *
     * @throws TransportConversionException if the xpath was not found
     *  in the given context.
     */
    public ParentNode adapt(ParentNode context, XPath path, T o) throws TransportConversionException {
        try {
            Node result = (Node) path.selectSingleNode(context);
            if(result == null)
                throw new IllegalArgumentException("XPath not found in document");
            if(result instanceof Attribute) {
                logger.info("Warning: XPath given pointed to an attribute. Using parent element instead.");
                result = result.getParent();
            }
            return adapt((ParentNode)result, o);
        } catch (JaxenException ex) {
            throw new TransportConversionException(ex.getMessage(), ex);
        }
    }

    /**
     * Unpacks by navigating to the standard starting position via XPath
     * and calling extract(Node, T).
     */
    public T unpack(Node data) throws TransportConversionException {
        return unpack(data, start);
    }

    /**
     * Unpacks by navigating to the specified XPath and calling
     * extract(Node, T).
     */
    public T unpack(Node data, XPath path) throws TransportConversionException {
        try {
            return converter.fromTransport(
                    splice.extract((Node) path.selectSingleNode(data),
                    transportPolicy), dataPolicy);
        } catch (JaxenException ex) {
            throw new TransportConversionException(ex.getMessage(), ex);
        }
    }

    /** 
     * Returns the name of the target. Defaults to the name of the class.
     */
    public String getName() {
        return getAdaptableClass().getName();
    }

    /** 
     * Return the target class. Must match T.
     */
    public Class<T> getAdaptableClass() {
        return adaptableClass;
    }

    public static <T> TransportAdapter<T> createStringAdapter(
            Document template, XPath start, TypeConverter<T, String> conv,
            ValuePolicy<T> dataPolicy, Class<T> clazz) {
        return new TemplateTransportAdapter<T, String>(template, start,
                new TextSplice(), conv, new NonEmptyString(),
                dataPolicy, clazz);
    }
    public static <T> TransportAdapter<T> createAttributeAdapter(
            Document template, XPath start, String attrName,
            TypeConverter<T, String> conv, ValuePolicy<T> dataPolicy, Class
            <T> clazz) {
        return new TemplateTransportAdapter<T, String>(template, start,
                new AttributeSplice(attrName), conv, new NonEmptyString(),
                dataPolicy, clazz);
    }

    public TransportAdapter<T> setContext(TransportContext context) {
        return this;
    }
}
