/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle, Sebastian Wrede, Jan Schaefer
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.xcf.fts;

import java.util.Collections;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Logger;

/**
 * Default implementation of the Event type which uses a hash-map to store
 * tags.
 *
 * @author iluetkeb
 */
public class DefaultEvent<T> implements Event<T> {
    private static final Logger logger =
            Logger.getLogger(DefaultEvent.class.getName());

    private final T data;
    private final ConcurrentMap<String, Object> tags;

    /** Create new event with empty tag set. */
    public DefaultEvent(T data) {
        this.data = data;
        tags = new ConcurrentHashMap<String, Object>(1);
    }

    protected DefaultEvent(T data, ConcurrentMap<String, Object> map) {
        this.data = data;
        this.tags = map;
    }

    @Override
    public T getData() {
        return data;
    }

    @Override
    public Object getTag(String name) {
        return tags.get(name);
    }

    protected final String makeName(String localName, String namespace) {
        return namespace + "." + localName;
    }

    @Override
    public Object getTag(String localName, String namespace) {
        return getTag(makeName(localName, namespace));
    }

    @Override
    public Map<String, Object> getTags() {
        return Collections.unmodifiableMap(tags);
    }

    /**
     * Create new event using the given data and the tags from the source event.
     *
     * @param data The data element
     * @param srcEvent The source event this new element was derived from.
     *   Tags from this event will be copied.
     * @param newTags List of new tag pairs to add. May be null.
     */
    public <K> Event<K> createDerived(K data) {
        return new DefaultEvent<K>(data, this.tags);
    }

    @Override
    public Event<T> addTags(Pair<String, Object> ... newTags) {
        logger.entering("DefaultEvent", "addTags", newTags);
        ConcurrentHashMap<String, Object> tagsCopy =
                new ConcurrentHashMap<String, Object>(this.tags);
        for(Pair<String, Object> tag : newTags)
            checkedPutTag(tagsCopy, tag.first, tag.second);
        return new DefaultEvent<T>(this.data, tagsCopy);
    }

    @Override
    public Event<T> removeTags(String ... names) {
        logger.entering("DefaultEvent", "removeTags", names);
        ConcurrentHashMap<String, Object> tagsCopy =
                new ConcurrentHashMap<String, Object>(this.tags);
        for(String name : names)
            tagsCopy.remove(name);
        return new DefaultEvent<T>(this.data, tagsCopy);
    }

    protected static final void checkedPutTag(ConcurrentMap<String, Object> map,
            String tagName, Object tagValue) {
        if(!putTag(map, tagName, tagValue))
                throw new IllegalStateException("Cannot set tag " + tagName +
                    " to '" + tagValue + "', it already has a value in " +
                    map);
    }

    protected static final boolean putTag(ConcurrentMap<String, Object> map,
            String tagName, Object tagValue) {
        if(tagName == null)
            throw new NullPointerException("Tag name must not be null");
        if(tagValue == null)
            throw new NullPointerException("Tag value for tag " + tagName +
                    " must not be null");
        return map.putIfAbsent(tagName, tagValue) == null;
    }

    @Override
    public String toString() {
        return "Event[data=" + data + ", tags=" + tags + "]";
    }

    public Event<T> mergeTags(Map<String, Object> otherTags) {
        logger.entering("DefaultEvent", "mergeTags", otherTags);
        ConcurrentHashMap<String, Object> tagsCopy =
                new ConcurrentHashMap<String, Object>(this.tags);
        for(Map.Entry<String, Object> tag : otherTags.entrySet())
            putTag(tagsCopy, tag.getKey(), tag.getValue());
        return new DefaultEvent<T>(this.data, tagsCopy);
    }

}
