/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle, Sebastian Wrede, Jan Schaefer
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.xcf.fts.engine;

import java.util.ArrayList;
import java.util.Deque;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.xcf.fts.Activatable;
import net.sf.xcf.fts.Event;
import net.sf.xcf.fts.Node;
import net.sf.xcf.fts.NullEvent;
import net.sf.xcf.fts.Triple;
import net.sf.xcf.fts.engine.EngineEventListener.EngineEvent;
import net.sf.xcf.fts.engine.EngineEventListener.EngineEventType;
import net.sf.xcf.fts.engine.jmx.EngineAgent;
import net.sf.xcf.fts.marks.EmitTypeMark;
import net.sf.xcf.fts.nodes.Slot;

/**
 * Base class for engine implementations that manages nodes and a queue. In
 * particular, it takes care of Activatables and initializers, queries sources
 * in a fair manner and
 *
 * Derived implementations need to provide a queuing strategy for new data.
 *
 * @author iluetkeb
 */
public abstract class AbstractEngine implements Engine, Runnable {

    private static final Logger logger = Logger.getLogger(AbstractEngine.class.getName());
    protected final Graph graph;
    // protected final SourceSelector selector;
    private ExecutorService sourcePollExecutor;
    protected ExecutorCompletionService<Triple<Node<?, ?>, Event<?>, Exception>> selector;
    protected final Deque<Triple<Node, Node, Event>> workQueue =
            new LinkedBlockingDeque<Triple<Node, Node, Event>>();
    private final Graph managementGraph = new Graph();
    private final Slot<NullEvent> nullSlot = new Slot<NullEvent>();
    private NodeErrorHandler errorHandler = new TerminatingNodeErrorHandler();
    protected int stepInterval = 0;
    // The listeners for core engine events
    private final List<EngineEventListener> engineListeners = new ArrayList<EngineEventListener>();

    // fairly coarse state management for the engine.
    // we could consider having more starting states, for the various steps,
    // but this simple one seems to suffice for now.
    enum EngineState {

        Idle,
        Starting, Running, Stopping,
    };
    private EngineState currentState = EngineState.Idle;
    private boolean jmxRegistered = false;

    /**
     * 
     * @param g  Graph, this graph has to contain connected Nodes already
     * @param fwMap  Map<Node, Forwarder>, the map Nodes matched with Forwarders
     * @throws NullPointerException  if no graph is set or the forwarder map is null
     */
    public AbstractEngine(Graph g) {
        if (g == null) {
            throw new NullPointerException("Graph must not be null");
        }
        logger.fine(System.identityHashCode(this) + " running graph " + g);
        managementGraph.mark(new EmitTypeMark<NullEvent>(nullSlot,
                NullEvent.class));

        this.graph = Graph.merge(g, managementGraph);
        registerJMX();
    }

    public void run() {
        registerJMX();

        synchronized (workQueue) {
            if (currentState != EngineState.Idle) {
                throw new IllegalStateException("Engine is not idle");
            }
            currentState = EngineState.Starting;
        }

        try {
            // startup
            if (!performActivatableAction(Activatable.Action.Start)) {
                return;
            }

            try {
                // initialize
                runInitializers();

                // execute graph until interrupted
                currentState = EngineState.Running;
                while (currentState == EngineState.Running
                        && !Thread.currentThread().isInterrupted()) {
                    runOnce();
                }
            } catch (TerminateEngineException ex) {
                logger.log(Level.WARNING, "Terminating due to exception {0}", ex);
            }

            // shutdown
            sourcePollExecutor.shutdownNow();
            try {
                if(!sourcePollExecutor.awaitTermination(500, TimeUnit.MILLISECONDS)) {
                    logger.log(Level.WARNING, "Source poller not terminated after 500ms");
                }
            } catch (InterruptedException ex) {
                // ignore
            }
            performActivatableAction(Activatable.Action.Stop);
        } finally {
            currentState = EngineState.Idle;
            sourcePollExecutor = null;
        }
    }

    EngineState getState() {
        return currentState;
    }

    private synchronized void registerJMX() {
        if (!jmxRegistered) {
            EngineAgent.getInstance().registerEngine(this, this.graph);
            jmxRegistered = true;
        }
    }

    private synchronized void unregisterJMX() {
        if (jmxRegistered) {
            EngineAgent.getInstance().unregisterEngine(this);
            jmxRegistered = false;
        }
    }

    /**
     * Submit source callables for all current sources.
     *
     * @return true if sources had not been watched before, false otherwise
     */
    boolean startSources() {
        if (sourcePollExecutor != null) {
            return false;
        }

        this.sourcePollExecutor = Executors.newFixedThreadPool(graph.getSourceNodes().size());
        this.selector = new ExecutorCompletionService<Triple<Node<?, ?>, Event<?>, Exception>>(sourcePollExecutor);
        for (Node<?, ?> n : graph.getSourceNodes()) {
            // do not use nullSlot as Source
            if (n == nullSlot) {
                continue;
            }
            logger.fine(System.identityHashCode(this) + " watching source " + n);
            SourceCompletionCallable c = new SourceCompletionCallable(this, n);
            this.selector.submit(c);
        }

        return true;
    }

    protected void runInitializers() throws TerminateEngineException {
        for (Node<?, ?> init : graph.getInitializerNodes()) {
            try {
                logger.log(Level.FINE, "Initializing {0}", init);
                addWorkItem(init, init.handleEvent(null));
            } catch (Exception ex) {
                fireNodeErrorEvent(null, init, null, ex);
            }
        }

    }

    public void shutdown() {
        requestStop();
        unregisterJMX();
    }

    @Override
    public void finalize() {
        shutdown();
    }

    @Override
    public void runOnce() {
        // I start the sources here instead of in run to
        // reduce incompatibility with the previous state where
        // they were started at construction time
        startSources();

        do {
            try {
                step();
                if (stepInterval > 0) {
                    Thread.sleep(this.stepInterval);
                }
            } catch (TerminateEngineException ex) {
                logger.logp(Level.SEVERE, getClass().getName(), "run",
                        "Error handler requested abort", ex);

                requestStop();
                return;
            } catch (InterruptedException ex) {
                requestStop();
                return;
            }
        } while (!workQueue.isEmpty());
    }

    private void requestStop() {
        synchronized (workQueue) {
            switch (currentState) {
                case Idle:
                    return;
                default:
                    currentState = EngineState.Stopping;
            }
        }
    }

    public void step() throws TerminateEngineException {
        logger.entering("AbstractEngine", "step", workQueue);
        if (workQueue.isEmpty()) {
            // work queue is empty, go back to sources
            try {
                Triple<Node<?, ?>, Event<?>, Exception> completed = this.selector.take().get();
                logger.fine(completed.toString());
                if (completed.third != null) {
                    fireNodeErrorEvent(null, completed.first, null, completed.third);
                }
                if (completed.second != null && completed.second.getData() != null) {
                    addWorkItem(completed.first, completed.second);
                }
                SourceCompletionCallable c = new SourceCompletionCallable(this, completed.first);
                this.selector.submit(c);
            } catch (InterruptedException e) {
                requestStop();
                return;
            } catch (RejectedExecutionException ex) {
                requestStop();
                return;
            } catch (ExecutionException e) {
                if (e.getCause() instanceof InterruptedException) {
                    requestStop();
                    return;
                }
                // this should not happen, as all exceptions are handled by the callable
                e.printStackTrace();
                assert (false);
            }

        } else {
            // take next item, if any, and process it
            Triple<Node, Node, Event> t = workQueue.poll();
            if (t == null) {
                return;
            }
            notifyEngineListeners(EngineEventType.WORK_ITEM_TAKEN, t.first, t.second, t.third.getData(), workQueue.size());

            try {
                final Event resultEvent = t.second.handleEvent(t.third);
                if (resultEvent != null && resultEvent.getData() != null) {
                    addWorkItem(t.second, resultEvent);
                } else {
                    if (graph.getEmergents(t.second).hasNext()
                            && !(t.third.getData() instanceof NullEvent)) {
                        addWorkItem(t.second, nullSlot,
                                t.third.createDerived(
                                new NullEvent(t.second, t.third)));
                    }
                }
            } catch (Exception ex) {
                fireNodeErrorEvent(t.first, t.second, t.third, ex);
            }

        }
        logger.exiting("AbstractEngine", "step", workQueue);
    }

    protected abstract void addWorkItem(Node source, Node next, Event input);

    @Override
    @SuppressWarnings("unchecked")
    public void addWorkItem(Node source, Event input) {
        logger.entering("AbstractEngine", "addWorkItem", input);
        if (input == null || input.getData() == null) {
            return;
        }

//        synchronized (workQueue) {
        if (input.getData() instanceof DestinationPair<?>) {
            final DestinationPair<?> dp = (DestinationPair<?>) (input.getData());
            addWorkItem(source, dp.first, input.createDerived(dp.second));
        } else if (input.getData() instanceof Iterator<?>) {
            for (Iterator<?> iter = (Iterator<?>) input.getData(); iter.hasNext(); /* in loop */) {
                addWorkItem(source, input.createDerived(iter.next()));
            }
        } else {
            for (Iterator<Node<?, ?>> i = graph.getEmergents(source); i.hasNext();) {
                Node<?, ?> dest = i.next();
                addWorkItem(source, dest, input);
                notifyEngineListeners(EngineEventType.WORK_ITEM_ADDED, source, dest, input, workQueue.size());
            }
        }
//        }
    }

    public NodeErrorHandler setErrorHandler(NodeErrorHandler h) {
        NodeErrorHandler prev = errorHandler;
        errorHandler = h;
        return prev;
    }

    protected void fireNodeErrorEvent(Node<?, ?> sourceNode,
            Node<?, ?> workNode, Event<? extends Object> data, Exception ex)
            throws TerminateEngineException {
        if (errorHandler == null) {
            return;
        }
        // we do not let error handlers interfere with interruptions
        if (ex instanceof InterruptedException) {
            requestStop();
            return;
        }
        @SuppressWarnings("unchecked")
        final NodeErrorEvent evt = new NodeErrorEvent(sourceNode, workNode, data, ex);
        try {
            errorHandler.eventHandlingFailed(evt);
        } catch (TerminateEngineException e) {
            throw e;
        } catch (Exception e) {
            logger.logp(Level.WARNING, AbstractEngine.class.getName(),
                    "fireNodeErrorEvent",
                    "caught exception during delivery of error notification",
                    e);
        }
    }

    public void setStepInterval(int itv) {
        this.stepInterval = itv;
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * net.sf.xcf.fts.engine.Engine#addEngineEventListener(net.sf.xcf.fts.engine
     * .AbstractEngine.EngineEventListener)
     */
    public void addEngineEventListener(EngineEventListener listener) {
        engineListeners.add(listener);
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * net.sf.xcf.fts.engine.Engine#removeEngineEventListener(net.sf.xcf.fts
     * .engine.AbstractEngine.EngineEventListener)
     */
    public boolean removeEngineEventListener(EngineEventListener listener) {
        return engineListeners.remove(listener);
    }

    /**
     * Notifies all the {@link EngineEventListener}s that an event has happened
     * in the engine.
     *
     * @param eventType
     *            The type of the event.
     * @param source
     *            A source node associated with this event.
     * @param dest
     *            A destination node associated with this event.
     * @param data
     *            A data object associated with this event.
     */
    protected void notifyEngineListeners(EngineEventType eventType,
            Node<?, ?> source, Node<?, ?> dest, Object data, int queueSize) {

        // If we have no listeners, return immediately
        if (engineListeners.isEmpty()) {
            return;
        }

        // Create the event
        EngineEvent event = new EngineEvent(eventType, source, dest, data, queueSize);

        notifyEngineListeners(event);
    }

    @Override
    public void notifyEngineListeners(EngineEvent event) {
        // Make a defensive copy so that listeners can remove themselves from
        // the list even when the event method is called.
        List<EngineEventListener> listenerCopy = new ArrayList<EngineEventListener>(
                engineListeners);

        // Notify all the listeners
        for (EngineEventListener listener : listenerCopy) {
            listener.handleEngineEvent(event);
        }
    }

    /**
     * Performs the given action on all activatables in the source
     * and sink node sets.
     */
    protected boolean performActivatableAction(Activatable.Action action) {
        boolean success = true;
        List<Node<?, ?>> nodes = new ArrayList<Node<?, ?>>();
        nodes.addAll(graph.getInitializerNodes());
        nodes.addAll(graph.getSourceNodes());
        nodes.addAll(graph.getSinkNodes());

        for (Node<?, ?> n : nodes) {
            if (n instanceof Activatable) {
                try {
                    logger.log(Level.FINE, "Setting state {0} on {1}", new Object[]{action, n});
                    ((Activatable) n).performAction(action);
                } catch (ExecutionException ex) {
                    logger.logp(Level.SEVERE, AbstractEngine.class.getName(),
                            "performActivatableAction", "Could not perform "
                            + action + ":" + ex.getMessage() + " on " + n,
                            ex.getCause());
                    success = false;
                }
            }
        }
        return success;
    }
}
