/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle, Sebastian Wrede, Jan Schaefer
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.xcf.fts.engine;

import net.sf.xcf.fts.Event;
import net.sf.xcf.fts.Node;
import net.sf.xcf.fts.engine.EngineEventListener.EngineEvent;

/**
 * The engine that handles all the communication in an FTS graph.
 * 
 * @author iluetkeb
 */
public interface Engine extends Runnable {
    /**
     * Add work item to this engines queue. The insertion point is
     * engine-specific.
     */
     void addWorkItem(Node<?,?> source, Event<? extends Object> input);

    /** 
     * Perform a single processing step. If no work items are available,
     * queries all sources.
     */
    void step() throws TerminateEngineException;

    /**
     * Query sources once, then run until there are no more workItems.
     */
    void runOnce();

    /**
     * Install handler for errors during event execution. May be null.
     * By default, the TerminatingErrorHandler is installed.
     * 
     * @return the previous handler, if any
     */
    NodeErrorHandler setErrorHandler(NodeErrorHandler handler);

    /**
     * Terminate all background processing by this engine. If the engine is
     * run in a separate thread, that thread should be interrupted prior
     * to this call, to ensure that background processing is not re-initiated.
     */
    void shutdown();
    
    /**
     * Adds a new listener that is notified on specific events inside the
     * engine.
     * 
     * @param listener
     *            The listener that should be added.
     * @see EngineEvent
     */
    void addEngineEventListener(EngineEventListener listener);

    /**
     * Tries to remove a lister from the list of engine event listeners.
     * 
     * @param listener
     *            The listener that should be removed.
     * @return <code>true</code>, if the listener was really removed.
     */
    boolean removeEngineEventListener(EngineEventListener listener);
    
    /**
     * Notifies all the {@link EngineEventListener}s that an event has happened
     * in the engine.
     * 
     * @param event The event that has happened.
     */
    void notifyEngineListeners(EngineEvent event);
}
