package net.sf.xcf.fts.engine;

import net.sf.xcf.fts.Node;
import net.sf.xcf.fts.engine.jmx.EngineEventWrapper;
import net.sf.xcf.fts.engine.jmx.NodeWrapper;

/**
 * A listener that can be notified about events that happen in the engine.
 * 
 * @author dklotz
 */
public interface EngineEventListener {

    /** Describes the type of the engine event. */
    public static enum EngineEventType {

        /** A new work item was added. */
        WORK_ITEM_ADDED,
        /** An item was taken from the work queue. */
        WORK_ITEM_TAKEN,
        HANDLE_EVENT_CALLED,
        HANDLE_EVENT_FINISHED
    }

    /**
     * Used to make some classes agnostic to the fact if they are working with
     * real {@link EngineEvent}s or {@link EngineEventWrapper}s.
     *
     * @param <NodeType> Should be either {@link Node} or {@link NodeWrapper}.
     */
    public static interface TimedNodeEvent<NodeType> {
        public long getCreationTime();
        public EngineEventType getType();
        public NodeType getSource();
    }

    /**
     * An event describing a change of state that happened inside the engine.
     *
     * @author dklotz
     */
    public static class EngineEvent implements TimedNodeEvent<Node<?,?>> {

        /** The type of this event. */
        public final EngineEventType type;
        /** When was this event created. */
        public final long creationTime = System.currentTimeMillis();
        /** A source node associated with this event. */
        public final Node<?, ?> source;
        /** A destination node associated with this event. */
        public final Node<?, ?> destination;
        /** A data object associated with this event. */
        public final Object data;
        /** The size of the work-queue or node-queue when this event was created. */
        public final int queueSize;

        /**
         * Creates a new engine event.
         *
         * @param type
         *            The type of the event, should never be null.
         * @param source
         *            A source node associated with this event.
         * @param dest
         *            A destination node associated with this event.
         * @param data
         *            A data object associated with this event.
         */
        public EngineEvent(EngineEventType type, Node<?, ?> source,
                Node<?, ?> dest, Object data) {
            this(type, source, dest, data, -1);
        }

        /**
         * Creates a new engine event.
         *
         * @param type
         *            The type of the event, should never be null.
         * @param source
         *            A source node associated with this event.
         * @param dest
         *            A destination node associated with this event.
         * @param data
         *            A data object associated with this event.
         * @param workQueueSize
         *            The current size of the work-queue.
         */
        public EngineEvent(EngineEventType type, Node<?, ?> source,
                Node<?, ?> dest, Object data, int workQueueSize) {
            this.type = type;
            this.source = source;
            this.destination = dest;
            this.data = data;
            this.queueSize = workQueueSize;

            assert (type != null);
        }

        public long getCreationTime() {
            return creationTime;
        }

        public Node getSource() {
            return source;
        }

        public EngineEventType getType() {
            return type;
        }
    }

    /**
     * React to an event in the engine. Don't block (or do anything compute
     * intensive) in this method, or you might slow down the core engine thread.
     *
     * @param event
     *            The event.
     */
    public void handleEngineEvent(EngineEvent event);
}
