/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle, Sebastian Wrede, Jan Schaefer
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.xcf.fts.engine;

import java.util.ArrayList;
import net.sf.xcf.fts.Node;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Logger;
import net.sf.xcf.fts.Pair;
import net.sf.xcf.fts.Sink;
import net.sf.xcf.fts.Source;
import net.sf.xcf.fts.engine.jmx.GraphWrapper;
import net.sf.xcf.fts.engine.jmx.NodeWrapper;
import net.sf.xcf.fts.marks.InitializerMark;
import net.sf.xcf.fts.marks.SourceConnectableMark;
import net.sf.xcf.fts.marks.Marked;
import net.sf.xcf.fts.marks.SinkMark;
import net.sf.xcf.fts.marks.SourceMark;
import net.sf.xcf.fts.nodes.filter.Filter;

/**
 * A Graph holds a number of Node instances and manages the connections
 * between them. Nodes may be added to and removed from the Graph directly or
 * by adding or removing a sub-Graph.
 * 
 * The graph may be observed for changes in its structure, thus allowing to use
 * it as a model for a processing setup. This allows to adapt other parts of a
 * data processing system to adapt according to the processing steps defined in
 * the graph.
 *
 * FIXME: The contents of incidents and emergents are confusing. They contain
 * "Edge" objects but once a is the source and once a is the sink of the edge.
 *
 * @author jschaefe
 * @author iluetkeb
 *
 */
public class Graph {

    private static final Logger logger = Logger.getLogger(Graph.class.getName());

    protected class Edge<T> {

        public final Node<?, ?> a;
        public final Node<?, ?> b;
        public AtomicInteger refcount = new AtomicInteger(0);

        public Edge(Node<?, ?> a, Node<?, ?> b) {
            this.a = a;
            this.b = b;
        }

        @Override
        public boolean equals(Object o) {
            if (!(o instanceof Graph.Edge<?>)) {
                return false;
            }
            return a.equals(((Edge<?>) o).a) && b.equals(((Edge<?>) o).b);
        }

        @Override
        public int hashCode() {
            int hash = 7;
            hash = 29 * hash + (this.a != null ? this.a.hashCode() : 0);
            hash = 29 * hash + (this.b != null ? this.b.hashCode() : 0);
            return hash;
        }

        @Override
        public String toString() {
            return "Edge[a=" + a + ", b=" + b + "]";
        }
    }
    /** Map for each node that contains a list of <em>outgoing</em> nodes. */
    private final ConcurrentMap<Node<?, ?>, CopyOnWriteArrayList<Edge<?>>> emergents =
            new ConcurrentHashMap<Node<?, ?>, CopyOnWriteArrayList<Edge<?>>>();
    /** Map for each node that contains a list of <em>incoming</em> nodes. */
    private final ConcurrentMap<Node<?, ?>, CopyOnWriteArrayList<Edge<?>>> incidents =
            new ConcurrentHashMap<Node<?, ?>, CopyOnWriteArrayList<Edge<?>>>();
    /** List of observers */
    protected final List<GraphObserver> observers = new LinkedList<GraphObserver>();
    /** Generic node markers for various processing purposes. */
    private Set<Marked> markers = new HashSet<Marked>(1);

    protected synchronized void addEdge(ConcurrentMap<Node<?, ?>, CopyOnWriteArrayList<Edge<?>>> map,
            Node<?, ?> a, Node<?, ?> b) {
        map.putIfAbsent(a, new CopyOnWriteArrayList<Edge<?>>());
        final List<Edge<?>> list = map.get(a);

        final Edge<?> e = new Edge(a, b);
        // TODO: use sorted list and binary search
        final int index = list.indexOf(e);
        if (index != -1) {
            list.get(index).refcount.incrementAndGet();
        } else {
            list.add(e);
        }
    }

    protected synchronized <T> void removeEdge(Map<Node<?, ?>, CopyOnWriteArrayList<Edge<?>>> map,
            Node<?, ? extends T> a, Node<? super T, ?> b) {
        final List<Edge<?>> list = map.get(a);
        if (list == null) {
            return;
        }
        final Edge<T> e = new Edge<T>(a, b);
        final int index = list.indexOf(e);
        if (index == -1) {
            return;
        }
        if (list.get(index).refcount.decrementAndGet() == 0) {
            list.remove(index);
            if (list.size() == 0) {
                map.remove(a);
            }
        }
    }

    /**
     * Add node mark.
     *
     * @return true when the mark was added (i.e., it was not yet present)
     * @see java.util.Set#add(java.lang.Object)
     */
    public boolean mark(Marked mark) {
        return this.markers.add(mark);
    }

    /**
     * Return all nodes that are marked with a given type of mark.
     */
    public Collection<Node<?,?>> markedWith(Class<? extends Marked> markType) {
        Set<Node<?,?>> marked = new HashSet<Node<?,?>>();
        for(Marked m : markers) {
            if(markType.isAssignableFrom(m.getClass()))
                marked.add(m.getNode());
        }
        return marked;
    }

    /**
     * Merge multiple graphs into one. The result graph will have the
     * super-set of all edges in the individual graphs. Additionally,
     * if nodes in the source graphs have been marked using a ConnectorMark,
     * the merge process will attempt to connect them.
     *
     * The input graphs will not be modified, but references to their Edges
     * and nodes will be placed in the new graph. All marks will also
     * be copied verbatim.
     *
     * @return the result graph
     */
    public static Graph merge(Graph... graphs) {
        Graph merged = new Graph();
        Set<SourceConnectableMark> connectors = new HashSet<SourceConnectableMark>();

        // copy all content        
        for (Graph g : graphs) {
            merged.addContent(g);
            findConnectors(connectors, g);
        }

        connect(connectors, merged);

        return merged;
    }

    protected void addContent(Graph g) {
        emergents.putAll(g.emergents);
        incidents.putAll(g.incidents);
        markers.addAll(g.markers);
    }

    void checkNodeUse(Node<?, ?> from, Node<?, ?> to) {
        if (from instanceof Sink<?>) {
            logger.warning("Sink node " + from + " specified as from node");
        }
        if (to instanceof Source<?>) {
            logger.warning("Source node " + to + " specified as to node");
        }

    }

    protected static void findConnectors(Set<SourceConnectableMark> connectors, Graph g) {
        for (Marked m : g.markers) {
            if (m instanceof SourceConnectableMark) {
                connectors.add((SourceConnectableMark) m);
            }
        }
    }

    protected static void connect(Set<SourceConnectableMark> connectors, Graph g) {
        for (SourceConnectableMark c : connectors) {
            match:
            for (Marked target : g.markers) {
                if (target == c) {
                    continue match;
                }
                c.connectIfCompatible(g, target);
            }
        }
    }

    /**
     * Merge single graph into this graph.
     * Note: if you have many graphs to add, use Graph.merge instead!
     */
    public Graph add(Graph g) {
        Set<SourceConnectableMark> connectors = new HashSet<SourceConnectableMark>(markers.size() + g.markers.size());
        addContent(g);
        //System.err.println(this);
        findConnectors(connectors, this);
        findConnectors(connectors, g);
        //System.err.println("connectors " + connectors);
        connect(connectors, this);
        return this;
    }

    public <T, U> GraphConnector<T, U> getConnector(Node<? super T, ? extends U> node) {
        return new GraphConnector<T, U>(this, node);
    }

    /**
     * Creates a connection between the two passed in nodes.
     * @param from
     * @param to
     */
    public <T, T2> GraphConnector<T, T2> connect(Node<?, ? extends T> from, Node<? super T, ? extends T2> to) {
        boolean fromNodeAdded = !(emergents.containsKey(from) || incidents.containsKey(from));
        boolean toNodeAdded = !(emergents.containsKey(to) || incidents.containsKey(to));
        addEdge(emergents, from, to);
        addEdge(incidents, to, from);
        notifyGraphObservers(from, to, true, fromNodeAdded, toNodeAdded);

        checkNodeUse(from, to);

        return new GraphConnector<T, T2>(this, to);
    }

    public <T> void mconnect(Node<?, ? extends Iterator<T>> from, Node<? super T, ?> to) {
        boolean fromNodeAdded = !(emergents.containsKey(from) || incidents.containsKey(from));
        boolean toNodeAdded = !(emergents.containsKey(to) || incidents.containsKey(to));
        addEdge(emergents, from, to);
        addEdge(incidents, to, from);

        checkNodeUse(from, to);

        notifyGraphObservers(from, to, true, fromNodeAdded, toNodeAdded);
    }

    public <T> void selectFirstOf(Node<?, ? extends T> from,
            Pair<? extends Filter<T>, Node<? super T, ?>>... cases) {
        FirstOfSelector selector = new FirstOfSelector(cases);
        connect(from, selector);
        for (Pair<? extends Filter<T>, Node<? super T, ?>> c : cases) {
            connect(selector, c.second);
        }
    }

    public <T, T2> void selectFirstOf(Node<?, ? extends T> from, Node<? super T2, ?> to,
            Pair<? extends Filter<T>, Node<? super T, ? extends T2>>... cases) {
        FirstOfSelector selector = new FirstOfSelector(cases);
        connect(from, selector);
        for (Pair<? extends Filter<T>, Node<? super T, ? extends T2>> c : cases) {
            connect(selector, c.second);
            connect(c.second, to);
        }
    }

    public <T> void selectAllOf(Node<?, ? extends T> from,
            Pair<? extends Filter<T>, Node<? super T, ?>>... cases) {
        AllOfSelector selector = new AllOfSelector(cases);
        connect(from, selector);
        for (Pair<? extends Filter<T>, Node<? super T, ?>> c : cases) {
            connect(selector, c.second);
        }
    }

    // did I mention that I can't stand generics? needless code-duplication...
    public <T, T2> void selectAllOf(Node<?, ? extends T> from, Node<? super T2, ?> to,
            Pair<? extends Filter<T>, Node<? super T, ? extends T2>>... cases) {
        AllOfSelector selector = new AllOfSelector(cases);
        connect(from, selector);
        for (Pair<? extends Filter<T>, Node<? super T, ? extends T2>> c : cases) {
            connect(selector, c.second);
            connect(c.second, to);
        }
    }

    /**
     * Breaks the connection between the two passed in nodes.
     * @param from
     * @param to
     */
    public void disconnect(Node from, Node to) {
        removeEdge(emergents, from, to);
        removeEdge(incidents, to, from);
        boolean fromNodeRemoved = !(emergents.containsKey(from) || incidents.containsKey(from));
        boolean toNodeRemoved = !(emergents.containsKey(to) || incidents.containsKey(to));
        notifyGraphObservers(from, to, false, fromNodeRemoved, toNodeRemoved);
    }

    private class NodeEdgeIterator implements Iterator<Node<?, ?>> {

        private final Iterator<Edge<?>> iter;

        public NodeEdgeIterator(Iterator<Edge<?>> iter) {
            this.iter = iter;
        }

        public boolean hasNext() {
            return iter.hasNext();
        }

        public Node<?, ?> next() {
            return iter.next().b;
        }

        public void remove() {
            throw new UnsupportedOperationException("Not supported yet.");
        }
    }
    private static final List<Node<?, ?>> emptyList = Collections.emptyList();

    protected Iterator<Node<?, ?>> getIterator(Map<Node<?, ?>, ? extends List<Edge<?>>> map, Node<?, ?> a) {
        if (a == null) {
            return emptyList.iterator();
        }

        final List<Edge<?>> e = map.get(a);
        if (e == null) {
            return emptyList.iterator();
        }
        return new NodeEdgeIterator(e.iterator());
    }

    public Iterator<Node<?, ?>> getEmergents(Node<?, ?> a) {
        return getIterator(emergents, a);
    }

    public Iterator<Node<?, ?>> getIncidents(Node<?, ?> a) {
        return getIterator(incidents, a);
    }

    /**
     * Checks if this {@link Node} is a source node (i.e. it has no incoming
     * edges).
     *
     * @param node
     *            The node to check (should be contained in this graph).
     * @return <code>true</code> if this node is a source node,
     *         <code>false</code> otherwise.
     */
    boolean isSource(Node<?, ?> node) {
        // If this node has no emerging edges, it can't be a source
        if (!emergents.containsKey(node)) {
            return false;
        }

        // otherwise, check if the incoming edges are initializers
        Set<Node<?, ?>> initializers = new HashSet<Node<?, ?>>(3);
        for (Marked m : markers) {
            if (m instanceof InitializerMark) {
                initializers.add(m.getNode());
            }
        }

        // initializers are not sources
        if (initializers.contains(node)) {
            return false;
        }

        // otherwise, if this node has no incoming edges, it is a source
        if (!incidents.containsKey(node)) {
            return true;
        }

        // its also a source if all of its incidents are initializers
        int initCount = 0;
        for (Edge<?> e : incidents.get(node)) {
            if (initializers.contains(e.b)) {
                ++initCount;
            }
        }

        return initCount >= incidents.get(node).size();
    }

    public Collection<Node<?, ?>> getInitializerNodes() {
        return markedWith(InitializerMark.class);
    }

    /**
     * Checks if this {@link Node} is a sink node (i.e. it has no outgoing
     * edges).
     *
     * @param node
     *            The node to check (should be contained in this graph).
     * @return <code>true</code> if this node is a sink node, <code>false</code>
     *         otherwise.
     */
    boolean isSink(Node<?, ?> node) {
        // If this node has no incoming edges, it can't be a sink
        if (!incidents.containsKey(node)) {
            return false;
        }

        // If this node has no outgoing edges, it is a sink
        if (!emergents.containsKey(node)) {
            return true;
        }

        // Otherwise, it isn't a sink
        return false;
    }

    /** Get Nodes without an incoming edge. */
    public Collection<Node<?, ?>> getSourceNodes() {
        Collection<Node<?, ?>> sources = markedWith(SourceMark.class);
        // backwards compat: if there are no explicit source marks, use isSource
        if(sources.isEmpty()) {
            for (Node<?, ?> n : emergents.keySet()) {
                if (isSource(n)) {
                    sources.add(n);
                }
            }
        }
        return sources;
    }

    /** Get Nodes without outgoing edges. */
    public Collection<Node<?, ?>> getSinkNodes() {
        Collection<Node<?, ?>> sinks = markedWith(SinkMark.class);
        // backwards compat: if there are no explicit sinks marks, use isSink
        for (Node<?, ?> n : incidents.keySet()) {
            if (isSink(n)) {
                sinks.add(n);
            }
        }
        return sinks;
    }

    public GraphWrapper wrap() {
        Set<Pair<NodeWrapper, NodeWrapper>> arcs = new HashSet<Pair<NodeWrapper, NodeWrapper>>();

        for (Collection<Edge<?>> edgeLists : emergents.values()) {
            for (Edge<?> edge : edgeLists) {
                NodeWrapper a = new NodeWrapper(edge.a);
                NodeWrapper b = new NodeWrapper(edge.b);

                arcs.add(Pair.create(a, b));
            }
        }

        Collection<Node<?, ?>> sources = getSourceNodes();
        List<NodeWrapper> sourceWraps = new ArrayList<NodeWrapper>(sources.size());

        for (Node<?, ?> source : sources) {
            sourceWraps.add(new NodeWrapper(source));
        }
        for(Node i : getInitializerNodes()) {
            sourceWraps.add(new NodeWrapper(i));
        }
        
        Collection<Node<?, ?>> sinks = getSinkNodes();
        List<NodeWrapper> sinkWraps = new ArrayList<NodeWrapper>(sinks.size());

        for (Node<?, ?> sink : sinks) {
            sinkWraps.add(new NodeWrapper(sink));
        }

        return new GraphWrapper(sourceWraps, sinkWraps, arcs);
    }

    public void addObserver(GraphObserver o) {
        if (!observers.contains(o)) {
            observers.add(o);
        }
    }

    public void removeObserver(GraphObserver o) {
        observers.remove(o);
    }

    @Override
    public String toString() {
        return "Graph[emergents=" + emergents + ", incidents=" + incidents
                + ", markers=" + markers + "]";
    }

    protected void notifyGraphObservers(Node<?, ?> from, Node<?, ?> to,
            boolean connect, boolean fromNodeChanged, boolean toNodeChanged) {
        // notify observers
        List<GraphObserver> copy = new LinkedList<GraphObserver>(observers);
        for (GraphObserver o : copy) {
            if (connect) {
                o.edgeAdded(from, to);
                if (fromNodeChanged) {
                    o.nodeAdded(from);
                }
                if (toNodeChanged) {
                    o.nodeAdded(to);
                }
            } else {
                o.edgeRemoved(from, to);
                if (fromNodeChanged) {
                    o.nodeRemoved(from);
                }
                if (toNodeChanged) {
                    o.nodeRemoved(to);
                }
            }
        }
    }

    /** Clear the graph of everything but observers. */
    public void clear() {
        this.emergents.clear();
        this.incidents.clear();
        this.markers.clear();
    }
}
