/*
 * GraphConnector.java
 *
 * Copyright (C) 2009 Bielefeld University
 * Copyright (C) 2009 Patrick Holthaus
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.xcf.fts.engine;

import net.sf.xcf.fts.marks.EmitTypeMark;
import net.sf.xcf.fts.marks.Mark;
import net.sf.xcf.fts.marks.ReceiveTypeMark;
import net.sf.xcf.fts.Node;

/**
 * This class may become handy to connet simple chains in a certain graph.
 * Marking as emit or receive type is also supported.
 * <br/>
 * Example usage:
 * {@code Graph g = new Graph().getConnector(Node<A,B>).connect(Node<B,C>).connect(Node<C,D>).emit(D).getGraph();}
 * <br/>
 * With the above code, {@code g} is a {@link Graph} containing three inter-connected
 * {@link Node}s (i.e. transforming data types {@code A->B->C->D}) of which the last one
 * is marked as an emitter for type D.
 * @param <T> In-Type of the associated {@link Node}
 * @param <U> Out-Type of the associated {@link Node}
 * @author pholthau
 */
public class GraphConnector<T, U> {

    private final Node<? super T, ? extends U> lastNode;
    private final Graph graph;

    /**
     * Creates a new GraphConnector for a certain {@link Graph}.
     * @param graph The {@link Graph} to create the connector for.
     * @param from The first {@link Node} that the next one should be connected
     * to.
     */
    public GraphConnector(Graph graph, Node<? super T, ? extends U> from){
        this.lastNode = from;
        this.graph = graph;
    }

    /**
     * Connects a new {@link Node} to the one associated with this connector
     * and returns another connector associated with the new {@link Node}.
     * @param <V> The Out-Type of the newly connected {@link Node}.
     * @param to The node to connect to.
     * @return A new {@link GraphConnector} associated with the newly connected {@link Node}.
     */
    public <V> GraphConnector<U, V> connect(Node<? super U, ? extends V> to){
        this.graph.connect(this.lastNode, to);
        return new GraphConnector<U, V>(this.graph, to);
    }

    /**
     * Marks the currently associated {@link Node} as an emitter in a specified scope.
     * @see EmitTypeMark
     * @param scope The emitting scope (makes use of {@link Mark#createScope(java.lang.String)}.
     * @param c The emitted class type.
     * @return s The {@link GraphConnector} for the currently associated {@link Node}.
     */
    public GraphConnector<T, U> emit(String scope, Class<? super U> c){
        this.graph.mark(new EmitTypeMark<U>(lastNode, Mark.createScope(scope), c));
        return this;
    }

    /**
     * Marks the currently associated {@link Node} as an emitter withouth any scope.
     * @see EmitTypeMark
     * @param c The emitted class type.
     * @return s The {@link GraphConnector} for the currently associated {@link Node}.
     */
    public GraphConnector<T, U> emit(Class<? super U> c){
        this.graph.mark(new EmitTypeMark<U>(lastNode, c));
        return this;
    }

    /**
     * Marks the currently associated {@link Node} as a receiver in a specified scope.
     * @see ReceiveTypeMark
     * @param scope The receiving scope (makes use of {@link Mark#createScope(java.lang.String)}.
     * @param c The received class type.
     * @return s The {@link GraphConnector} for the currently associated {@link Node}.
     */
    public GraphConnector<T, U> receive(String scope, Class<? extends T> c){
        this.graph.mark(new ReceiveTypeMark<T>(lastNode, Mark.createScope(scope), c));
        return this;
    }

    /**
     * Marks the currently associated {@link Node} as a receiver without a scope.
     * @see ReceiveTypeMark
     * @param c The received class type.
     * @return s The {@link GraphConnector} for the currently associated {@link Node}.
     */
    public GraphConnector<T, U> receive(Class<? extends T> c){
        this.graph.mark(new ReceiveTypeMark<T>(lastNode, c));
        return this;
    }

    /**
     * Returns a copy of the associated {@link Graph}.
     * @return a new {@link Graph} that contains every connection made with
     * this {@link GraphConnector} so far.
     */
    public Graph getGraph(){
        return new Graph().add(this.graph);
    }
}