/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle <ingo@fargonauten.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.xcf.fts.engine;

import java.lang.annotation.Annotation;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.sf.xcf.fts.Event;
import net.sf.xcf.fts.Node;
import net.sf.xcf.fts.annotations.ReceiveLast;
import net.sf.xcf.fts.engine.EngineEventListener.EngineEventType;

/**
 * A very simple queueing engine that uses one thread per node, a shared
 * queue for outgoing work items and one in-queue per node.
 *
 * TODO convert this (or implement another class) to use an ExecutorService
 * for handling node execution, so that threads may be allocated more
 * efficiently. the main thing to watch out for is to prevent one node
 * from getting all the threads.
 *
 * @author iluetkeb
 */
public class QueueingEngine extends AbstractEngine {

    private static final Logger logger = Logger.getLogger(QueueingEngine.class.getName());
    private final ConcurrentMap<Node<?, ?>, NodeWorker<?, ?>> workers =
            new ConcurrentHashMap<Node<?, ?>, NodeWorker<?, ?>>();

    private static final AtomicInteger serial = new AtomicInteger();
    private final ThreadGroup tg = new ThreadGroup("QEngineGroup-" + serial.getAndIncrement());

    public QueueingEngine(Graph graph) {
        super(graph);
        logger.entering(getClass().getName(), "init");
        tg.setDaemon(true);
    }

    @Override
    @SuppressWarnings("unchecked")
    protected void addWorkItem(Node source, Node next, Event input) {
        final Event evt = input;
        if(evt == null || evt.getData() == null)
            return;
        logger.entering("QueueingEngine", "addWorkItem");
        if(logger.isLoggable(Level.FINER)) {
            logger.finer("source=" + source + ", next=" + next + ", input=" + input);
        }
        try {
            NodeWorker<?, ?> targetWorker = workers.get(next);
            // if worker is missing, add new one
            if (targetWorker == null) {
                targetWorker = new NodeWorker(next);
                NodeWorker nw = workers.putIfAbsent(next, targetWorker);
                if(nw != null)
                    targetWorker = nw;
                else
                    targetWorker.start();
            }
            targetWorker.put(evt);
        } catch (InterruptedException ex) {
            Logger.getLogger(QueueingEngine.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * Invokes super-class shutdown and then interrupts all running
     * workers.
     */
    @Override
    public void shutdown() {
        super.shutdown();
        for (NodeWorker<?, ?> worker : workers.values()) {
            worker.interrupt();
        }
    }

    private class NodeWorker<T1, T2> extends Thread {

        private final Node<T1, T2> node;
        private BlockingQueue<Event<? extends T1>> inQueue = new LinkedBlockingQueue<Event<? extends T1>>();
        private final boolean receiveLast;

        public NodeWorker(Node<T1, T2> node) {
            super(tg, node.toString());
            this.node = node;
            receiveLast = this.node.getClass().getAnnotation(ReceiveLast.class) != null;
            if(receiveLast)
                logger.info("Node " + node + " configured for receiveLast");
        }

        @Override
        public void run() {
            while (!interrupted()) {
                try {
                    Event<? extends T1> evt = inQueue.take();
                    
                    // Let the node handle the event
                    notifyEngineListeners(EngineEventType.HANDLE_EVENT_CALLED, node, null, null, inQueue.size());
                    Event<? extends Object> result = node.handleEvent(evt);
                    notifyEngineListeners(EngineEventType.HANDLE_EVENT_FINISHED, node, null, null, inQueue.size());
                    
                    addWorkItem(node, result);
                } catch (InterruptedException ex) {
                    return;
                } catch (Exception ex) {
                    logger.logp(Level.WARNING, "QueueingEngine#NodeThread", "run",
                            "Exception on node execution", ex);
                    return;
                }

            }
        }

        public void put(Event<? extends T1> evt) throws InterruptedException {
            if(receiveLast)
                inQueue.clear();
            
            inQueue.put(evt);

            if(inQueue.size() > 100) {
                logger.warning("Inqueue for " + node + " larger than 100.");
            }
        }
    }
}
