/**
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2009 David Klotz <mail -at- planetraven -dot- de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.xcf.fts.engine.jmx;

import java.lang.management.ManagementFactory;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.NotCompliantMBeanException;
import javax.management.ObjectName;

import net.sf.xcf.fts.engine.Engine;
import net.sf.xcf.fts.engine.EngineEventListener;
import net.sf.xcf.fts.engine.Graph;

/**
 * A JMX agent that enables monitoring and management of a running FTS
 * {@link Engine}.
 * 
 * @author dklotz
 */
public class EngineAgent {

    private static final Logger logger = Logger.getLogger(EngineAgent.class
            .getName());
    private static EngineAgent instance;

    private final MBeanServer mbs;
    private final AtomicLong counter = new AtomicLong();
    private final Map<Engine, Long> engineIds = Collections
            .synchronizedMap(new HashMap<Engine, Long>());

    /**
     * Creates the agent. Private because this is a singleton class that should
     * only be instantiated by itself.
     */
    private EngineAgent() {
        mbs = ManagementFactory.getPlatformMBeanServer();
    }

    /**
     * Get (and, if necessary, create) the singleton instance of this class.
     * 
     * @return The single instance of this class, never null.
     */
    public static synchronized EngineAgent getInstance() {
        if (instance == null) {
            instance = new EngineAgent();
        }

        return instance;
    }

    /**
     * Registers a running FTS engine with this agent. This creates the
     * necessary MBean and registers it with the MBean server. The MBean is
     * registered under a name like the following:
     * 
     * <pre>
     * net.sf.xcf.fts.engine.jmx:name=Engine,id=number
     * </pre>
     * 
     * 
     * @param engine
     *            The engine (must not be null).
     * @param graph
     *            The graph this engine is working on (must also not be null).
     */
    public void registerEngine(Engine engine, Graph graph) {
        long id = counter.getAndIncrement();

        try {
            ObjectName engineName = createEngineName(id);

            EngineInfoMBean engineBean = new EngineInfo(id, engine, graph);

            engineIds.put(engine, id);
            mbs.registerMBean(engineBean, engineName);

            // register global listeners
            for(EngineEventListener l : globalListeners) {
                engine.addEngineEventListener(l);
            }

        } catch (InstanceAlreadyExistsException e) {
            // Since we use a unique id, this should not happen
            logger.log(Level.SEVERE,
                    "Could not register engine MBean, mbean exists!", e);
        } catch (MBeanRegistrationException e) {
            // I don't know when this could happen, but it shouldn't
            logger.log(Level.SEVERE, "Could not register engine MBean!", e);
        } catch (NotCompliantMBeanException e) {
            // This should not happen, normally
            logger.log(Level.SEVERE,
                    "Could not register engine MBean, mbean not compliant!", e);
        }
    }

    /**
     * Unregisters a FTS engine from this agent. This unregisters the associated
     * MBean from the MBean server.
     * 
     * @param engine
     *            An engine that should already have been registered.
     */
    public void unregisterEngine(Engine engine) {
        if (!engineIds.containsKey(engine)) {
            logger
                    .warning("Trying to unregister engine that was not registered!");
            return;
        }

        long id = engineIds.get(engine);

        try {
            ObjectName engineName = createEngineName(id);
            mbs.unregisterMBean(engineName);
            // remove global listeners
            for(EngineEventListener l : globalListeners) {
                engine.removeEngineEventListener(l);
            }
            engineIds.remove(engine);

        } catch (MBeanRegistrationException e) {
            // This happens when we were interrupted in the meantime
            logger.log(Level.FINE, "Could not unregister engine MBean!", e);
        } catch (InstanceNotFoundException e) {
            // This should not happen, since we checked first
            logger.log(Level.SEVERE,
                    "Could not unregister engine MBean, mbean not found!", e);
        }
    }

    /**
     * 
     * @param id
     * @return
     */
    private ObjectName createEngineName(long id) {
        ObjectName engineName = null;

        try {
            engineName = new ObjectName(
                    "net.sf.xcf.fts.engine.jmx:name=Engine,id=" + id);
        } catch (MalformedObjectNameException e) {
            // This should not happen, the name is OK
            logger.log(Level.SEVERE,
                    "Could not create name for an engine MBean!", e);
        }

        return engineName;
    }

    private List<EngineEventListener> globalListeners = new
            CopyOnWriteArrayList<EngineEventListener>();

    public boolean isRegistered(Engine e) {
        return engineIds.containsKey(e);
    }

    /**
     * Add global engine listener. The listener will be added
     * to all existing engines and all that are subsequently
     * created.
     */
    public void addGlobalListener(EngineEventListener l) {
        globalListeners.add(l);
        for(Engine e : this.engineIds.keySet()) {
            e.addEngineEventListener(l);
        }
    }
    /**
     * Remove global engine listener. Will detach from all
     * current engines.
     */
    public void removeGlobalListener(EngineEventListener l) {
        globalListeners.remove(l);
        for(Engine e : this.engineIds.keySet()) {
            e.removeEngineEventListener(l);
        }
    }
    
}
