/**
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2009 David Klotz <mail -at- planetraven -dot- de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.xcf.fts.engine.jmx;

import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicLong;

import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.NotificationBroadcasterSupport;

import net.sf.xcf.fts.Node;
import net.sf.xcf.fts.engine.Engine;
import net.sf.xcf.fts.engine.EngineEventListener;
import net.sf.xcf.fts.engine.Graph;

/**
 * The implementation of an MBean that provides monitoring and management
 * capabilities for an FTS engine.
 * 
 * @author dklotz
 * @see EngineInfoMBean
 */
public class EngineInfo extends NotificationBroadcasterSupport implements EngineInfoMBean, EngineEventListener {

    private final Engine engine;
    private final Graph originalGraph;
    private GraphWrapper wrappedGraph;
    private final long id;
    private final AtomicLong seq = new AtomicLong();
    private MBeanNotificationInfo[] notifications;
    private int lastKnownWorkQueueSize = -1;
    private ProcessingTimeHelper<Node<?, ?>> processingTimeHelper = new ProcessingTimeHelper<Node<?,?>>();

    public EngineInfo(final long id, Engine engine, Graph graph) {
        //super(Executors.newCachedThreadPool());
        super(Executors.newFixedThreadPool(3));

        this.id = id;
        this.engine = engine;
        this.originalGraph = graph;

        this.engine.addEngineEventListener(this);
    }

    @Override
    public long getId() {
        return id;
    }

    @Override
    public GraphWrapper getGraph() {
        if (wrappedGraph == null) {
            wrappedGraph = originalGraph.wrap();
        }
        return wrappedGraph;
    }

    @Override
    public void handleEngineEvent(EngineEvent event) {
        if (event.type == EngineEventType.WORK_ITEM_ADDED
                || event.type == EngineEventType.WORK_ITEM_TAKEN) {
            lastKnownWorkQueueSize = event.queueSize;
        }

        if (event.type == EngineEventType.HANDLE_EVENT_CALLED
                || event.type == EngineEventType.HANDLE_EVENT_FINISHED) {
            updateAverageProcessingTime(event);
        }

        Notification eventNotification = new Notification(
                EngineInfoMBean.FTS_NOTIFICATION_TYPE, this, seq
                        .getAndIncrement());

        eventNotification.setTimeStamp(event.creationTime);
        EngineEventWrapper eventWrapper = new EngineEventWrapper(event);
        eventNotification.setUserData(eventWrapper);

        // Note that this happens asynchronously, since we supplied an Executor
        // in the constructor. This does not wait for the listener to complete.
        sendNotification(eventNotification);
    }

    private void updateAverageProcessingTime(EngineEvent event) {
        processingTimeHelper.eventHappened(event);
    }

    @Override
    public synchronized MBeanNotificationInfo[] getNotificationInfo() {
        if (notifications == null) {
            String[] types = new String[] { EngineInfoMBean.FTS_NOTIFICATION_TYPE };

            String name = Notification.class.getName();

            String description = "An engine event has happened.";

            MBeanNotificationInfo info = new MBeanNotificationInfo(types, name,
                    description);

            notifications = new MBeanNotificationInfo[] { info };
        }

        return notifications;
    }

    @Override
    public int getWorkQueueSize() {
        return lastKnownWorkQueueSize;
    }

    @Override
    public double getAverageProcessingTime() {
        return processingTimeHelper.getAverageProcessingTime();
    }
}
