/**
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2009 David Klotz <mail -at- planetraven -dot- de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.xcf.fts.engine.jmx;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Collection;

import javax.management.MBeanServerConnection;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;

import net.sf.xcf.fts.Pair;
import net.sf.xcf.fts.export.DotGraphExport;

/**
 * A JMX client that connects to a running FTS application, gets its graphs and
 * exports them using the {@link DotGraphExport}.
 * 
 * The application that will be monitored needs to be started with parameters
 * like this:
 * 
 * <pre>
 * -Dcom.sun.management.jmxremote.port=9999 
 * -Dcom.sun.management.jmxremote.authenticate=false 
 * -Dcom.sun.management.jmxremote.ssl=false
 * </pre>
 * 
 * @author dklotz
 */
public class FTSJMXGraphExport {

    // Print an error and quit
    private static void fatalError(String reason) {
        System.err.println(reason);
        System.exit(1);
    }

    /**
     * Connects to an FTS application via JMX and exports its graphs using the
     * {@link DotGraphExport}.
     * 
     * @param args
     *            If we have two arguments, they are taken as "host" and "port",
     *            otherwise they are ignored and default values (localhost:9999)
     *            are used.
     */
    public static void main(String[] args) {
        // Default hostname and port
        String hostname = "localhost";
        String port = "9999";

        // If we have exactly two commandline arguments, take them as host/port
        if (args.length == 2) {
            hostname = args[0];
            port = args[1];
        }

        DotGraphExport dotGrapher = new DotGraphExport();

        try {
            // Connect to JMX on the given host/port
            JMXServiceURL url = new JMXServiceURL(
                    "service:jmx:rmi:///jndi/rmi://" + hostname + ":" + port
                            + "/jmxrmi");

            JMXConnector jmxc = JMXConnectorFactory.connect(url);

            // Get the MBean connector
            MBeanServerConnection mbsc = jmxc.getMBeanServerConnection();

            // Query for our MBeans
            Collection<Pair<ObjectName, EngineInfoMBean>> mbeans = JMXUtils
                    .getRegisteredMBeans(mbsc);

            // If we didn't find any, quit with an error
            if (mbeans.isEmpty()) {
                fatalError("No FTS engine MBeans found on " + hostname + ":"
                        + port);
            }

            // Get the graph from all the MBeans...
            for (Pair<ObjectName, EngineInfoMBean> mbean : mbeans) {
                // ... and add it to the DotGrapher
                dotGrapher.addGraph("engine" + mbean.second.getId(),
                        mbean.second.getGraph());
            }

            // Close the JMX connection
            jmxc.close();

            // Print the dot graph output
            System.out.println(dotGrapher.getDotString());

        } catch (MalformedURLException e) {
            fatalError("Could not find JMX service: " + e.getMessage());
        } catch (IOException e) {
            fatalError("Error talking to the JMX service: " + e.getMessage());
        }
    }

}
