/**
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2009 David Klotz <mail -at- planetraven -dot- de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.xcf.fts.engine.jmx;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import java.util.Set;
import net.sf.xcf.fts.Node;
import net.sf.xcf.fts.Pair;
import net.sf.xcf.fts.engine.Graph;
import nu.xom.Attribute;
import nu.xom.Element;
import nu.xom.Elements;

/**
 * A wrapper that encapsulates all the relevant information about a
 * {@link Graph} for sending it via JMX. It can be seen as a
 * {@link Serializable} subset of the information contained in a {@link Graph}.
 *
 * @author dklotz
 */
public class GraphWrapper implements Serializable {
    private static final String GRAPH_ELEMENT_NAME = "ftsGraph";
    private static final String NODES_ELEMENT_NAME = "nodes";
    private static final String ARCS_ELEMENT_NAME = "arcs";
    private static final String ARC_ELEMENT_NAME = "arc";
    private static final String FROM_ATTRIBUTE_NAME = "from";
    private static final String SINK_ATTRIBUTE_NAME = "sink";
    private static final String SOURCE_ATTRIBUTE_NAME = "source";
    private static final String TO_ATTRIBUTE_NAME = "to";

    private final List<NodeWrapper> sourceNodes;
    private final List<NodeWrapper> sinkNodes;
    private Set<NodeWrapper> allNodes = new HashSet<NodeWrapper>();
    private Map<NodeWrapper, Collection<NodeWrapper>> emergents = new HashMap<NodeWrapper, Collection<NodeWrapper>>();
    private Map<NodeWrapper, Collection<NodeWrapper>> incidents = new HashMap<NodeWrapper, Collection<NodeWrapper>>();
    /** To enable serialization. */
    private static final long serialVersionUID = -4929854315501386842L;

    /**
     * <em>Deprecated:</em> Use {@link Graph#wrap()} instead.
     *
     * @param graph The graph that should be wrapped.
     * @deprecated Use {@link Graph#wrap()}.
     */
    @Deprecated
    public GraphWrapper(Graph graph) {
        Collection<Node<?, ?>> sources = graph.getSourceNodes();
        Collection<Node<?, ?>> sinks = graph.getSinkNodes();

        GraphWrapper wrapped = graph.wrap();

        this.sourceNodes = wrapped.sourceNodes;
        this.sinkNodes = wrapped.sinkNodes;
        this.allNodes = wrapped.allNodes;
        this.emergents = wrapped.emergents;
        this.incidents = wrapped.incidents;

        assert (sourceNodes.size() == sources.size());
        assert (sinkNodes.size() == sinks.size());
    }

    /**
     * Creates a new graph wrapper from a list of arcs / edges.
     *
     * @param sources The source nodes.
     * @param sinks The sink nodes.
     * @param arcs All the arcs in the graph.
     */
    public GraphWrapper(List<NodeWrapper> sources, List<NodeWrapper> sinks, Collection<Pair<NodeWrapper, NodeWrapper>> arcs) {
        this.sourceNodes = sources;
        this.sinkNodes = sinks;

        // Create the emergent-map-entries for all the nodes
        for (Pair<NodeWrapper, NodeWrapper> arc : arcs) {
            allNodes.add(arc.first);
            allNodes.add(arc.second);
        }

        for (NodeWrapper node : allNodes) {
            emergents.put(node, new ArrayList<NodeWrapper>());
            incidents.put(node, new ArrayList<NodeWrapper>());
        }

        // Add the emergents
        for (Pair<NodeWrapper, NodeWrapper> arc : arcs) {
            emergents.get(arc.first).add(arc.second);
            incidents.get(arc.second).add(arc.first);
        }
    }

    Set<NodeWrapper> getAllNodes() {
        return allNodes;
    }

    public Collection<NodeWrapper> getEmergents(NodeWrapper node) {
        return emergents.get(node);
    }
    public Collection<NodeWrapper> getIncidents(NodeWrapper node) {
        return incidents.get(node);
    }

    public Collection<NodeWrapper> getSourceNodes() {
        return sourceNodes;
    }

    public Collection<NodeWrapper> getSinkNodes() {
        return sinkNodes;
    }

    public boolean isSource(NodeWrapper source) {
        return sourceNodes.contains(source);
    }

    public boolean isSink(NodeWrapper source) {
        return emergents.get(source).isEmpty();
    }

    /**
     * Converts this graph to an XML element.
     * @return A XOM {@link Element}.
     */
    public Element toElement() {
        Element graphRoot = new Element(GRAPH_ELEMENT_NAME);

        // Nodes
        Element nodes = new Element("nodes");

        for (NodeWrapper node : getAllNodes()) {
            Element nodeElement = node.toElement();
            
            String sourceAttributeValue = Boolean.FALSE.toString();
            if (getSourceNodes().contains(node)) {
                sourceAttributeValue = Boolean.TRUE.toString();
            }
            nodeElement.addAttribute(new Attribute(SOURCE_ATTRIBUTE_NAME, sourceAttributeValue));

            String sinkAttributeValue = Boolean.FALSE.toString();
            if (getSinkNodes().contains(node)) {
                sinkAttributeValue = Boolean.TRUE.toString();
            }
            nodeElement.addAttribute(new Attribute(SINK_ATTRIBUTE_NAME, sinkAttributeValue));

            nodes.appendChild(nodeElement);
        }

        graphRoot.appendChild(nodes);

        // Arcs
        Element arcs = new Element("arcs");
        for (Map.Entry<NodeWrapper, Collection<NodeWrapper>> emergent : emergents.entrySet()) {
            int fromId = emergent.getKey().getId();
            for (NodeWrapper toNode : emergent.getValue()) {
                Element arc = new Element("arc");
                arc.addAttribute(new Attribute("from", String.valueOf(fromId)));
                arc.addAttribute(new Attribute("to", String.valueOf(toNode.getId())));
                arcs.appendChild(arc);
            }
        }
        graphRoot.appendChild(arcs);

        return graphRoot;
    }

    public static GraphWrapper fromXML(Element graphRoot) {
        // TODO: Or a query for the first <ftsGraph> child using XPath? Slower?
        if (!GRAPH_ELEMENT_NAME.equals(graphRoot.getLocalName())) {
            throw new IllegalArgumentException("The XML root element for a graph must be <" + GRAPH_ELEMENT_NAME + ">");
        }

        // For the sources and sinks
        List<NodeWrapper> sources = new ArrayList<NodeWrapper>();
        List<NodeWrapper> sinks = new ArrayList<NodeWrapper>();

        // Parse the nodes
        Element nodesElement = graphRoot.getFirstChildElement(NODES_ELEMENT_NAME);
        Elements nodeElements = nodesElement.getChildElements(NodeWrapper.NODE_ELEMENT_NAME);
        List<NodeWrapper> nodes = new ArrayList<NodeWrapper>(nodeElements.size());
        Map<Integer, NodeWrapper> idMap = new HashMap<Integer, NodeWrapper>(nodeElements.size());

        for (int i = 0; i < nodeElements.size(); i++) {
            Element nodeElement = nodeElements.get(i);

            NodeWrapper node = NodeWrapper.fromXML(nodeElement);

            if (Boolean.parseBoolean(nodeElement.getAttributeValue(SOURCE_ATTRIBUTE_NAME))) {
                sources.add(node);
            }

            if (Boolean.parseBoolean(nodeElement.getAttributeValue(SINK_ATTRIBUTE_NAME))) {
                sinks.add(node);
            }

            nodes.add(node);
            idMap.put(node.getId(), node);
        }

        // Parse the arcs
        Element arcsElement = graphRoot.getFirstChildElement(ARCS_ELEMENT_NAME);
        Elements arcElements = arcsElement.getChildElements(ARC_ELEMENT_NAME);
        List<Pair<NodeWrapper, NodeWrapper>> arcs = new ArrayList<Pair<NodeWrapper, NodeWrapper>>();

        for (int i = 0; i < arcElements.size(); i++) {
            Element arcElement = arcElements.get(i);
            int fromId = Integer.parseInt(arcElement.getAttributeValue(FROM_ATTRIBUTE_NAME));
            int toId = Integer.parseInt(arcElement.getAttributeValue(TO_ATTRIBUTE_NAME));

            arcs.add(Pair.create(idMap.get(fromId), idMap.get(toId)));
        }

        return new GraphWrapper(sources, sinks, arcs);
        //return new GraphWrapper(nodes, arcs);
    }
}
