/**
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2009 David Klotz <mail -at- planetraven -dot- de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.xcf.fts.engine.jmx;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import javax.management.InstanceNotFoundException;
import javax.management.JMX;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanServerConnection;
import javax.management.MalformedObjectNameException;
import javax.management.Notification;
import javax.management.NotificationListener;
import javax.management.ObjectName;

import net.sf.xcf.fts.Pair;

/**
 * Some utility methods for working with the FTS JMX MBeans.
 * 
 * @author dklotz
 */
public class JMXUtils {

    private static final Logger logger = Logger.getLogger(JMXUtils.class
            .getName());

    /**
     * Get all the {@link EngineInfoMBean}s and the names they are registered
     * under.
     * 
     * @param mbsc
     *            The MBean connection, not null.
     * @return A collection with {@link Pair}s of the MBeand and its name, never
     *         null, but may be empty.
     * @throws IOException
     *             A communication problem occurred when talking to the MBean
     *             server.
     */
    public static Collection<Pair<ObjectName, EngineInfoMBean>> getRegisteredMBeans(
            MBeanServerConnection mbsc) throws IOException {

        List<Pair<ObjectName, EngineInfoMBean>> beans;

        try {
            // Query for MBeans matching our name pattern
            ObjectName mbeanQuery = new ObjectName(
                    "net.sf.xcf.fts.engine.jmx:name=Engine,id=?");
            Set<ObjectName> beanNames = mbsc.queryNames(mbeanQuery, null);

            beans = new ArrayList<Pair<ObjectName, EngineInfoMBean>>(beanNames
                    .size());

            for (ObjectName beanName : beanNames) {
                EngineInfoMBean mbeanProxy = JMX.newMBeanProxy(mbsc, beanName,
                        EngineInfoMBean.class, true);
                beans.add(new Pair<ObjectName, EngineInfoMBean>(beanName,
                        mbeanProxy));
            }

        } catch (MalformedObjectNameException e) {
            // This should not happen, as long as the name above is valid
            throw new IllegalStateException("Illegal JMX object name?!", e);
        }

        return beans;
    }

    /**
     * Registers an {@link EngineEventHandler} to receive notifications from the
     * {@link EngineInfoMBean}.
     * 
     * @param mbsc
     *            The MBean connection, not null.
     * @param beanName
     *            The name of the MBean, not null.
     * @param handler
     *            The handler that receives event notifications, not null.
     * @return A {@link Runnable} object that can (and should) be used to
     *         unregister the listener when done.
     * @throws InstanceNotFoundException
     *             The MBean name provided does not match any of the registered
     *             MBeans.
     * @throws IOException
     *             A communication problem occurred when talking to the MBean
     *             server.
     */
    public static Runnable connectEngineEventHandler(
            final MBeanServerConnection mbsc, final ObjectName beanName,
            final EngineEventHandler handler) throws InstanceNotFoundException,
            IOException {

        // Create a listener that calls the handler when appropriate
        final NotificationListener listener = new NotificationListener() {

            @Override
            public void handleNotification(Notification notification,
                    Object handback) {
                if (EngineInfoMBean.FTS_NOTIFICATION_TYPE.equals(notification
                        .getType())) {

                    handler.handleEvent((EngineEventWrapper) notification
                            .getUserData());
                }
            }
        };

        // Register that listener for notifications from the MBean
        mbsc.addNotificationListener(beanName, listener, null, null);

        // Create a runnable that can be used to remove that listener fromm all
        // notifications
        Runnable removeRunnable = new Runnable() {

            @Override
            public void run() {
                try {
                    mbsc.removeNotificationListener(beanName, listener);
                } catch (InstanceNotFoundException e) {
                    // This should not happen
                    logger
                            .warning("Could not remove the event visualizer JMX listener, MBean not found!");
                } catch (ListenerNotFoundException e) {
                    // This should not happen
                    logger
                            .warning("Could not remove the event visualizer JMX listener, listener not found!");
                } catch (IOException e) {
                    // I don't really know how we could handle this case
                    logger
                            .warning("Could not remove the event visualizer JMX listener, could not talk to MBean server!");
                }
            }
        };

        return removeRunnable;
    }
}
