/**
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2009 David Klotz <mail -at- planetraven -dot- de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.xcf.fts.engine.jmx;

import java.io.Serializable;
import java.lang.annotation.Annotation;
import java.lang.reflect.Method;

import net.sf.xcf.fts.Node;
import net.sf.xcf.fts.Sink;
import net.sf.xcf.fts.Source;
import net.sf.xcf.fts.nodes.AbstractNode;
import net.sf.xcf.fts.nodes.filter.Filter;
import net.sf.xcf.fts.nodes.transform.AbstractTransformNode;
import nu.xom.Attribute;
import nu.xom.Element;

/**
 * A wrapper that encapsulates all the relevant information about a {@link Node}
 * for sending it via JMX. It is basically a {@link Serializable} subset of the
 * information contained in a {@link Node}.
 *
 * @author dklotz
 */
public class NodeWrapper implements Serializable {
    private static final String CLASS_ATTRIBUTE_NAME = "class";
    private static final String DEFAULT_TYPE_VALUE = "node";
    private static final String FILTER_TYPE_VALUE = "filter";
    private static final String ID_ATTRIBUTE_NAME = "id";
    private static final String NAME_ATTRIBUTE_NAME = "name";
    static final String NODE_ELEMENT_NAME = "node";
    private static final String TRANSFORM_TYPE_VALUE = "transform";
    private static final String TYPE_ATTRIBUTE_NAME = "type";

    /** To enable serialization. */
    private static final long serialVersionUID = 5287370835769665893L;
    private final int id;
    private final String name;
    private final String originalToString;
    private final String originalClass;
    private final boolean transformNode;
    private final boolean filterNode;
    private final boolean isSource;
    private final boolean isSink;
    private final Annotation[] annotations;

    public NodeWrapper(Node<?, ?> node) {
        this.id = System.identityHashCode(node);

        // check source/sink status
        if(node instanceof Source) {
            isSource = true;
            isSink = false;
        } else  {
            isSource = false;
            if(node instanceof Sink) {
                isSink = true;
            } else {
                isSink = false;
            }
        }

        // Determine the type of the node
        if (AbstractTransformNode.class.isAssignableFrom(node.getClass())) {
            this.transformNode = true;
        } else {
            this.transformNode = false;
        }

        // copy annotations
        annotations = node.getClass().getAnnotations();


        if (Filter.class.isAssignableFrom(node.getClass())) {
            this.filterNode = true;
        } else {
            this.filterNode = false;
        }

        // Get some information about the original implementation
        this.originalClass = node.getClass().getName();
        this.originalToString = node.toString();

        // Get the name (or, alternatively, the toString value) of the node
        if (node instanceof AbstractNode<?, ?>) {
            AbstractNode<?, ?> an = (AbstractNode<?, ?>) node;
            this.name = an.getName();
        } else {
            this.name = node.toString();
        }
    }

    protected NodeWrapper(int id, boolean transformNode, boolean filterNode,
            String originalClass, String originalToString, String name,
            boolean isSink, boolean isSource, Annotation a[]) {
        this.id = id;
        this.transformNode = transformNode;
        this.filterNode = filterNode;
        this.originalClass = originalClass;
        this.originalToString = originalToString;
        this.name = name;
        this.isSink = isSink;
        this.isSource = isSource;
        this.annotations = a;
    }

    protected NodeWrapper(int id, boolean transformNode, boolean filterNode, String originalClass, String originalToString, String name) {
        this(id, transformNode, filterNode, originalClass, originalToString, name, false, false, null);
    }

    /**
     * Returns the object id of the original node this wrapper is based on.
     *
     * @return A unique int for each node instance in the same JVM, can be
     *         negative.
     * @see System#identityHashCode(Object)
     */
    public int getId() {
        return id;
    }

    /**
     * Returns a name for this node. For nodes extending from
     * {@link AbstractNode}, this is the name property. For all other nodes this
     * is simply the value of the nodes toString method.
     *
     * @return A String, can only be null if the original node returned null.
     * @see #getOriginalToString()
     */
    public String getName() {
        return name;
    }

    /**
     * Checks whether a node is a transform node. That is, whether it is derived
     * from the class {@link AbstractTransformNode}. Nodes that do not derive
     * from that class will not be considered transform nodes by this method.
     *
     * @return <code>true</code> if the node derives from
     *         {@link AbstractTransformNode}, <code>false</code> otherwise.
     * @see #isFilterNode()
     */
    public boolean isTransformNode() {
        return transformNode;
    }

    /**
     * Checks whether a node is a filter node. That is, whether it is derived
     * from something implementing {@link Filter}. Nodes that do not implement
     * that interface will not be considered filter nodes by this method.
     *
     * @return <code>true</code> if the original node implemented {@link Filter}
     *         , <code>false</code> otherwise.
     * @see #isTransformNode()
     */
    public boolean isFilterNode() {
        return filterNode;
    }

    public boolean isSink() {
        return isSink;
    }
    public boolean isSource() {
        return isSource;
    }

    public boolean isAnnotationPresent(Class<? extends Annotation> requestedAnnotation) {
        for(Annotation a : annotations) {
            if(requestedAnnotation.isAssignableFrom(a.annotationType()))
                return true;
        }
        return false;
    }

    /**
     * Returns the name of the original nodes class.
     *
     * @return A String, not null.
     * @see Class#getName()
     */
    public String getOriginalClass() {
        return originalClass;
    }

    /**
     * Returns the unchanged value of the original nodes toString method.
     *
     * @return The unchanged return value of the original toString, can only be
     *         null if that method returned null.
     */
    public String getOriginalToString() {
        return originalToString;
    }

    @Override
    public String toString() {
        return "[Node: id=" + id + ", name=" + name + "]";
    }

    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof NodeWrapper)) {
            return false;
        }

        NodeWrapper other = (NodeWrapper) obj;
        return (id == other.id) && name.equals(other.name);
    }

    @Override
    public int hashCode() {
        return id + 23 * name.hashCode();
    }

    /**
     * Converts this node to an XML element.
     * @return A XOM {@link Element}.
     */
    public Element toElement() {
        Element nodeElement = new Element(NODE_ELEMENT_NAME);

        // Attributes:
        nodeElement.addAttribute(new Attribute(ID_ATTRIBUTE_NAME, String.valueOf(getId())));
        nodeElement.addAttribute(new Attribute(NAME_ATTRIBUTE_NAME, getName()));
        nodeElement.addAttribute(new Attribute(CLASS_ATTRIBUTE_NAME, getOriginalClass()));
        String typeString = DEFAULT_TYPE_VALUE;
        if (isTransformNode()) {
            typeString = TRANSFORM_TYPE_VALUE;
        } else if (isFilterNode()) {
            typeString = FILTER_TYPE_VALUE;
        }
        nodeElement.addAttribute(new Attribute( TYPE_ATTRIBUTE_NAME, typeString));

        // The elements value
        nodeElement.appendChild(getOriginalToString());

        return nodeElement;
    }

    public static NodeWrapper fromXML(Element nodeRoot) {
        // TODO: Or a query for the first <node> child using XPath? Slower?
        if (!NODE_ELEMENT_NAME.equals(nodeRoot.getLocalName())) {
            throw new IllegalArgumentException("The XML root element for a node must be <" + NODE_ELEMENT_NAME + ">");
        }

        // The elements value
        String originalToString = nodeRoot.getValue();

        // Attributes
        int id = Integer.parseInt(nodeRoot.getAttributeValue(ID_ATTRIBUTE_NAME));
        String name = nodeRoot.getAttributeValue(NAME_ATTRIBUTE_NAME);
        String className = nodeRoot.getAttributeValue(CLASS_ATTRIBUTE_NAME);

        String type = nodeRoot.getAttributeValue(TYPE_ATTRIBUTE_NAME);
        boolean isTransformNode = TRANSFORM_TYPE_VALUE.equals(type);
        boolean isFilterNode = FILTER_TYPE_VALUE.equals(type);

        return new NodeWrapper(id, isTransformNode, isFilterNode, className, originalToString, name);
    }
}
