/**
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2009 David Klotz <mail -at- planetraven -dot- de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.xcf.fts.engine.jmx;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.logging.Logger;

import net.sf.xcf.fts.Node;
import net.sf.xcf.fts.engine.EngineEventListener.EngineEventType;
import net.sf.xcf.fts.engine.EngineEventListener.TimedNodeEvent;

/**
 * A class that helps calculating the average processing time spent in a node.
 * 
 * @author dklotz
 * 
 * @param <NodeType>
 *            The type for the node classes, probably {@link Node} or
 *            {@link NodeWrapper}.
 */
public class ProcessingTimeHelper<NodeType> {

    private static final Logger logger = Logger.getLogger(ProcessingTimeHelper.class.getName());
    private final Map<NodeType, Long> hitCounts = new HashMap<NodeType, Long>();
    private final Map<NodeType, Long> lastBeginTimes = new HashMap<NodeType, Long>();
    private final Map<NodeType, Double> averageProcessingTimes = new HashMap<NodeType, Double>();
    private final Thread queueEater;
    private final BlockingQueue<TimedNodeEvent<NodeType>> incomingEvents = new LinkedBlockingQueue<TimedNodeEvent<NodeType>>();

    public ProcessingTimeHelper() {
        queueEater = new Thread(new Runnable() {

            public void run() {
                processIncoming();
            }
        });
        queueEater.setDaemon(true);
        queueEater.start();
    }

    @Override
    protected void finalize() throws Throwable {
        queueEater.interrupt();
    }

    public void eventHappened(TimedNodeEvent<NodeType> event) {
        if (!incomingEvents.offer(event)) {
            logger.severe("Queue overflow in the ProcessingTimeHelper!");
        }
    }

    private void processIncoming() {
        while (!Thread.currentThread().isInterrupted()) {
            try {
                TimedNodeEvent<NodeType> event = incomingEvents.take();

                if (event.getType() == EngineEventType.HANDLE_EVENT_CALLED) {
                    handleEventCalled(event);
                } else if (event.getType() == EngineEventType.HANDLE_EVENT_FINISHED) {
                    handleEventFinished(event);
                }
            } catch (InterruptedException e) {
                // Exit the loop
                break;
            }
        }
    }

    private void handleEventCalled(TimedNodeEvent<NodeType> event) {
        lastBeginTimes.put(event.getSource(), event.getCreationTime());
    }

    private void handleEventFinished(TimedNodeEvent<NodeType> event) {
        NodeType node = event.getSource();
        long timeStamp = event.getCreationTime();

        if (!lastBeginTimes.containsKey(node)) {
            logger.warning("handleEvent finished without being called?");
            return;
        } else if (lastBeginTimes.get(node) > timeStamp) {
            logger.warning("handleEvent finished before it was called?");
            return;
        }

        long lastBeginTime = lastBeginTimes.get(node);
        long processingTime = timeStamp - lastBeginTime;

        assert (processingTime >= 0) : "Negative processing time?!";

        long hitCount = 0;
        if (hitCounts.containsKey(node)) {
            hitCount = hitCounts.get(node);
        }

        double lastAverage = 0.0;

        synchronized (averageProcessingTimes) {
            if (averageProcessingTimes.containsKey(node)) {
                lastAverage = averageProcessingTimes.get(node);
            }

            averageProcessingTimes.put(node, incrementalAverage(hitCount,
                    lastAverage, processingTime));
        }

    }

    /**
     * Calculates the average processing time over all the known nodes that have
     * processed <b>at least one event</b>. Nodes that have not handled any
     * event (yet) are not taken into account.
     * 
     * @return A positive double, or 0.0.
     */
    public double getAverageProcessingTime() {
        double totalAvg = 0.0;

        synchronized (averageProcessingTimes) {
            for (double avg : averageProcessingTimes.values()) {
                totalAvg += avg;
            }

            totalAvg /= averageProcessingTimes.values().size();
        }

        return totalAvg;
    }

    /**
     * Get the average processing time spent in the handleEvent method of a
     * node. If the never was never seen before, this method returns -1.0.
     * 
     * @param node
     *            The node, not null.
     * @return A positive double or -1.0.
     */
    public double getAverageProcessingTime(NodeType node) {
        if (!averageProcessingTimes.containsKey(node)) {
            return -1;
        }

        return averageProcessingTimes.get(node);
    }

    private static double incrementalAverage(long steps, double oldAverage,
            double newValue) {

        double oldFactor;
        double newFactor;

        if (steps > 0) {
            oldFactor = ((steps - 1.0) / steps);
            newFactor = (1.0 / steps);
        } else {
            oldFactor = 0.0;
            newFactor = 1.0;
        }

        return (oldFactor * oldAverage) + (newFactor * newValue);
    }
}
