/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package net.sf.xcf.fts.export;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import net.sf.xcf.fts.annotations.Fuse;

import net.sf.xcf.fts.engine.Graph;
import net.sf.xcf.fts.engine.jmx.GraphWrapper;
import net.sf.xcf.fts.engine.jmx.NodeWrapper;

/**
 * Experiment to produce dot output that looks like a UML activity graph.
 * The object will keep
 * track of graph nodes already seen and won't add them again (this
 * is also a protection against endless loops for graphs with circles).
 * Multiple graphs may be added and they will all appear in the output.
 *
 * @author iluetkeb
 * @author dklotz
 */
public class ActivityGraphExport {
    protected Set<NodeWrapper> nodesSeen = new HashSet<NodeWrapper>();
    private final StringBuilder dotDoc = new StringBuilder();

    private static final Map<String, String> barAttrs;
    private static final Map<String, String> diamondAttrs;
    static {
        HashMap<String, String> attrs = new HashMap<String, String>();
        attrs.put("label", "\"\"");
        attrs.put("shape", "rect");
        attrs.put("style", "filled");
        attrs.put("fillcolor", "black");
        attrs.put("height", "0.1");
        barAttrs = Collections.unmodifiableMap(attrs);
        
        attrs = new HashMap<String, String>();
        attrs.put("label", "\"\"");
        attrs.put("style", "\"\"");
        attrs.put("shape", "diamond");
        diamondAttrs = Collections.unmodifiableMap(attrs);
    }

    protected void appendDescendants(StringBuilder sb, GraphWrapper g, NodeWrapper n) {
        if(nodesSeen.contains(n))
            return;
        nodesSeen.add(n);

        Map<String, String> nodeAttrs = new HashMap<String, String>();
        Map<String, String> edgeAttrs = new HashMap<String, String>();

        if(n.isFilterNode()) {
            nodeAttrs.putAll(diamondAttrs);
            edgeAttrs.put("taillabel", "\"[" + formatNodeLabel(n) + "]\"");
        } else if(n.isAnnotationPresent(Fuse.class)) {
            nodeAttrs.putAll(barAttrs);
            String nameParts[] = n.getOriginalClass().split("\\.");
            edgeAttrs.put("taillabel", "\"" + nameParts[nameParts.length -1] + "\"");
        } else {
            nodeAttrs.put("shape", "box");
            nodeAttrs.put("style", "rounded");
            nodeAttrs.put("label", "\"" + formatNodeLabel(n) + "\"");
        }

        String id = Integer.toString(n.getId());

        Collection<NodeWrapper> incidents = g.getIncidents(n);
        // multiple inputs and no fusion -> insert join node
        if(incidents.size() > 1 && !n.isFilterNode() && !n.isAnnotationPresent(Fuse.class)) {
            sb.append("// insert fuse before ").append(formatNodeLabel(n)).append("\n");
            sb.append(id);
            sb.append(formatAttrs(diamondAttrs));
            sb.append(id + " -> ");
            id = "real" + id;
            sb.append(id).append("\n");
        }

        sb.append(id);
        sb.append(formatAttrs(nodeAttrs));


        Collection<NodeWrapper> emergents = g.getEmergents(n);
        // introduce splitting node, if necessary
        if(emergents.size() > 1) {
            sb.append("// inserting split after ").append(formatNodeLabel(n)).append("\n");;
            String splitid = "split" + id;
            sb.append(id).append(" -> ").append(splitid);
            sb.append(formatAttrs(edgeAttrs));
            sb.append(splitid).append(formatAttrs(barAttrs));

            // change data for following
            edgeAttrs.clear();
            id = splitid;
        }
        
        for(NodeWrapper next: emergents) {
            sb.append(id).append(" -> ");
            sb.append(next.getId());
            sb.append(formatAttrs(edgeAttrs));
            appendDescendants(sb, g, next);
        }

        if(n.isSource()) {
            sb.append("start -> ").append(n.getId()).append("\n");
        } else if(n.isSink()) {
            sb.append(n.getId()).append(" -> end\n");
        }
    }

    protected String formatAttrs(Map<String, String> attrs) {
        if(attrs.size() == 0) {
            return ";\n";
        }
        StringBuilder sb = new StringBuilder();
        sb.append(" [");

        String[] parts = new String[attrs.size()];
        int index = 0;
        for(Map.Entry<String, String> e : attrs.entrySet()) {
            parts[index++] = e.getKey() + '=' + e.getValue();
        }
        for(int i = 0; i < parts.length; ++i) {
            sb.append(parts[i]);
            if(i < (parts.length -1))
                sb.append(",");
        }


        return sb.append("];\n").toString();
    }

    protected String formatNodeLabel(NodeWrapper n) {
        return n.getOriginalToString().replace("\"", "\\\"");
    }

    public void addGraph(String name, Graph g) {
        addGraph(name, new GraphWrapper(g));
    }
    
    public void addGraph(String name, GraphWrapper g) {
        nodesSeen.clear();
        dotDoc.append("\ndigraph ").append(name).append(" {\n");
        dotDoc.append("splines=false\n");
        dotDoc.append("start[shape=circle, style=filled];\n");
        dotDoc.append("end[shape=doublecircle, style=filled];\n");
        // traverse all edges and add them
        Collection<NodeWrapper> sources = g.getSourceNodes();
        for(NodeWrapper n : sources) {
            appendDescendants(dotDoc, g, n);
        }
        dotDoc.append("}\n");
    }

    public String getDotString() {
        return dotDoc.toString();
    }
}
