/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle <ingo@fargonauten.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.xcf.fts.nodes;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.SynchronousQueue;
import net.sf.xcf.fts.nodes.source.AbstractSource;
import net.sf.xcf.fts.Event;
import net.sf.xcf.fts.Sink;
import net.sf.xcf.fts.Source;

/**
 * Bridges two graphs, transporting full events (not just event data)
 * between them.
 *
 * @author iluetkeb
 */
public class EventQueueBridge<T> {
    private final BlockingQueue<Event<? extends T>> queue;
    private final Source<T> source = new AbstractSource<T>("Bridge-Source") {
        @Override
        public Event<? extends T> next() throws Exception {
            return queue.take();
        }
    };
    private final Sink<T> sink = new Sink<T>() {
        public void push(Event<? extends T> e) throws Exception {
            queue.add(e);
        }
        public Event<? extends Void> handleEvent(Event<? extends T> input) throws Exception {
            push(input);
            return null;
        }
        @Override
        public String toString() {
            return "Bridge-Sink";
        }
    };

    EventQueueBridge(BlockingQueue<Event<? extends T>> queue) {
        this.queue = queue;
    }

    public Sink<T> getSink() {
        return sink;
    }
    public Source<T> getSource() {
        return source;
    }

    /**
     * Creates a bridge that only passes on an element when there is a thread
     * waiting to receive it.
     *
     * @param <T> event content type
     * @param clazz The type of the event-contents -- needed just for type inference
     * @return The new bridge.
     */
    public static <T> EventQueueBridge<T> createSynchronousBridge(Class<T> clazz) {
        return new EventQueueBridge<T>(new SynchronousQueue<Event<? extends T>>());
    }
    /**
     * Creates a bridge with a maximum capacity (elements waiting to be 
     * retrieved by the source. Putting elements into the sink after the 
     * capacity has been reached will cause an IllegalStateException.
     * 
     * @param capacity The maximum capacity of the queue 
     * @param clazz The type of the event-contents -- needed just for type inference
     * @return The new bridge.
     */
    public static <T> EventQueueBridge<T> createLimitedBridge(int capacity, Class<T> clazz) {
        return new EventQueueBridge<T>(new ArrayBlockingQueue<Event<? extends T>>(capacity));
    }
    /**
     * Creates a bridge with a maximum capacity of Integer.MAX_VALUE, which
     * is the maximum size an array may have. This is as unlimited as it gets
     * ;-)
     *
     * @param clazz The type of the event-contents -- needed just for type inference
     * @return The new bridge.
     */
    public static <T> EventQueueBridge<T> createUnlimitedBridge(Class<T> clazz) {
        return new EventQueueBridge<T>(new LinkedBlockingQueue<Event<? extends T>>());
    }
}
