/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle <ingo@fargonauten.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.xcf.fts.nodes;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.logging.Logger;
import net.sf.xcf.fts.Event;
import net.sf.xcf.fts.Node;
import net.sf.xcf.fts.nodes.filter.AbstractFilter;

/**
 * A helper class that provides two filters for synchronizing event flow through
 * two graphs. The filters will block until the synchronizer condition in
 * the other graph has been reached. The graph that is beginning the
 * synchronization is the "leading" graph and uses the leading filter. The
 * other graph is the "follower" graph and uses the following filter.
 *
 * @param <TL> Type for the filter in <b>L</b>eading graph
 * @param <TF> Type for the filter in <b>F</b>ollowing graph
 *
 * @author iluetkeb
 */
public class GraphSynchronizer<TL,TF> {
    private static final Logger logger =
            Logger.getLogger(GraphSynchronizer.class.getName());
    /**
     * Synchronization queue.
     */
    private final BlockingQueue<Integer> queue;
    
    private final Node<TL,TL> leadFilter = new AbstractFilter<TL>() {
        @Override
        public boolean isMatch(Event e) throws Exception {
            logger.entering("GraphSynchronizer", "leader", e);
            // NOTE: do not put e, as it would create a reference
            queue.put(1);
            logger.exiting("GraphSynchronizer", "leader", e);
            return true;
        }
        @Override
        public String toString() {
            return "leader";
        }
    };
    private final Node<TF,TF> followFilter = new AbstractFilter<TF>() {
        @Override
        public boolean isMatch(Event<? extends TF> e) throws Exception {
            logger.entering("GraphSynchronizer", "follower", e);
            queue.take();
            logger.exiting("GraphSynchronizer", "follower", e);
            return true;
        }
        @Override
        public String toString() {
            return "follower";
        }
    };

    GraphSynchronizer(BlockingQueue<Integer> queue) {
        this.queue = queue;
    }

    /**
     * Filter for use in the leading graph.
     *
     * @return the leading graph filter (always the same instance)
     */
    public Node<TL,TL> getLeadingFilter() {
        return leadFilter;
    }
    /**
     * Filter for use in the following graph.
     *
     * @return the following graph filter (always the same instance)
     */
    public Node<TF,TF> getFollowingFilter() {
        return followFilter;
    }

    /**
     * Creates a synchronizer that will block when the putting graph
     * has put "limit" items more than the taking graph has taken. This
     * is, e.g. usefull for protocols where only limit commands may be
     * written
     *
     * @param <TL> Content type for the leading filter
     * @param <TF> Content type for the following filter
     * @param limit The maximum number of outstanding objects.
     * @param class1 marker class instance for putting type
     * @param class2 marker class instance for taking type
     * @return
     */
    public static <TL,TF> GraphSynchronizer<TL,TF>
            createLimitSynchronizer(int limit, Class<TL> class1, Class<TF> class2) {
        return new GraphSynchronizer<TL,TF>(new ArrayBlockingQueue<Integer>(limit));
    }
}
