/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle <ingo@fargonauten.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.xcf.fts.nodes.correlation;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.TimeUnit;
import net.sf.xcf.fts.Event;

import net.sf.xcf.fts.Node;

/**
 * Correlates using the source tag added by a TaggingEventfactory. Example:
 * <code>
 *  Graph g;
 *  InjectorQueueHeadSource source;
 *  Node someTransform;
 *  SourceTagCorrelator sink;
 *  g.connect(source, someTransform);
 *  g.connect(someTransform, sink);
 *
 *  Event e = source.inject(data);
 *  Object result = sink.waitForResult(e);
 *
 * </code>
 *
 * handleEvent of this class blocks until a correlatable has been retrieved.
 *
 * @see net.sf.xcf.fts.TaggingEventFactory
 * @see net.sf.xcf.fts.nodes.source.InjectorQueueHeadSource
 * @author iluetkeb
 */
public abstract class AbstractCorrelator<T> implements Node<T,Void>, OutputCorrelator<T> {

    protected abstract BlockingQueue<Event<? extends T>> createNewQueue();
    protected abstract BlockingQueue<Event<? extends T>> getMatchingQueue(Event<?> evt);
    protected abstract void removeMatchingQueue(Event<?> evt);

    @Override
    public final Event<? extends Void> handleEvent(Event<? extends T> e) throws Exception {
        getMatchingQueue(e).put(e);
        return null;
    }

    public final T waitForResult(Event<?> in) throws InterruptedException {
        final T result = getMatchingQueue(in).take().getData();
        removeMatchingQueue(in);
        return result;
    }

    public final T waitForResult(Event<?> in, long timeout, TimeUnit unit) throws InterruptedException {
        final Event<? extends T> evt = getMatchingQueue(in).poll(timeout, unit);
        if(evt != null) {
            removeMatchingQueue(in);
            return evt.getData();
        }
        return null;
    }
}
