/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle <ingo@fargonauten.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.xcf.fts.nodes.correlation;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.TimeUnit;
import net.sf.xcf.fts.Event;

import net.sf.xcf.fts.Node;
import static net.sf.xcf.fts.TaggingEventFactory.TAG_KEY;

/**
 * Correlates using the source tag added by a TaggingEventfactory. Example:
 * <code>
 *  Graph g;
 *  InjectorQueueHeadSource source;
 *  Node someTransform;
 *  SourceTagCorrelator sink;
 *  g.connect(source, someTransform);
 *  g.connect(someTransform, sink);
 *
 *  Event e = source.inject(data);
 *  Object result = sink.waitForResult(e);
 *
 * </code>
 *
 * handleEvent of this class blocks until a correlatable has been retrieved.
 *
 * @see net.sf.xcf.fts.TaggingEventFactory
 * @see net.sf.xcf.fts.nodes.source.InjectorQueueHeadSource
 * @author iluetkeb
 */
public abstract class SourceTagCorrelator<T> extends AbstractCorrelator<T> {


    private final ConcurrentMap<Object, BlockingQueue<Event<? extends T>>>
        correlationMap = new ConcurrentHashMap<Object, BlockingQueue<Event<? extends T>>>();

    @Override
    protected BlockingQueue<Event<? extends T>> getMatchingQueue(Event<?> evt) {
        final Object tag = evt.getTag(TAG_KEY);
        if(tag == null)
            throw new IllegalArgumentException("Event has no " + TAG_KEY + " tag");
        correlationMap.putIfAbsent(tag, createNewQueue());
        return correlationMap.get(tag);
    }

    @Override
    protected void removeMatchingQueue(Event<?> evt) {
        correlationMap.remove(evt.getTag(TAG_KEY));
    }

    /**
     * Create a source tag correlator that will block the threading providing
     * the correlating event upon entering handleEvent, until waitForResult
     * is called with a matching event.
     */
    public static <T> SourceTagCorrelator<T> createBlockingCorrelator(Class<T> clazz) {
        return new SourceTagCorrelator<T>() {
            @Override
            protected BlockingQueue<Event<? extends T>> createNewQueue() {
                return new SynchronousQueue<Event<? extends T>>();
            }
        };
    }

    /**
     * Create a source tag correlator that will not block the thread providing
     * the correlating event (but waitForResult will block as usual).
     */
    public static <T> SourceTagCorrelator<T> createNonblockingCorrelator(Class<T> clazz) {
        return new SourceTagCorrelator<T>() {
            @Override
            protected BlockingQueue<Event<? extends T>> createNewQueue() {
                return new ArrayBlockingQueue<Event<? extends T>>(1);
            }
        };
    }
}
