/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle <ingo@fargonauten.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.xcf.fts.nodes.fusion;

import java.util.UUID;
import net.sf.xcf.fts.Event;
import net.sf.xcf.fts.Pair;
import net.sf.xcf.fts.annotations.Fuse;
import net.sf.xcf.fts.annotations.NodeName;
import net.sf.xcf.fts.nodes.AbstractNode;
import net.sf.xcf.fts.nodes.transform.TaggingNode;

/**
 *
 * @author iluetkeb
 */
@NodeName("MakePair") @Fuse
public class PairFusion<T1,T2> extends AbstractNode<Object,Pair<T1,T2>> {
    protected Event<? extends T1> first;
    protected Event<? extends T2> second;

    private static final String TAG_KEY = TaggedFusionNode.MAP_FUSE_NAME;

    private final UUID uuid1, uuid2;
    private final TaggingNode tagFirst;
    private final TaggingNode tagSecond;

    /** Use this to tag data from the first source. */
    public TaggingNode getTagFirst() {
        return tagFirst;
    }
    /** Use this to tag data from the second source. */
    public TaggingNode getTagSecond() {
        return tagSecond;
    }

    public PairFusion() {
        this(UUID.randomUUID(), UUID.randomUUID());
    }
    public PairFusion(UUID uuid1, UUID uuid2) {
        if(uuid1.equals(uuid2))
            throw new IllegalArgumentException("UUIDs must be distinct " +
                    "(got 1st=" + uuid1 + ", 2nd=" + uuid2 + ")");
        this.uuid1 = uuid1;
        this.uuid2 = uuid2;
        tagFirst = new TaggingNode(TAG_KEY, uuid1);
        tagSecond = new TaggingNode(TAG_KEY, uuid2);
    }
    public PairFusion(String one, String two) {
        this(UUID.nameUUIDFromBytes(one.getBytes()), UUID.nameUUIDFromBytes(two.getBytes()));
    }
    
    public Event<? extends Pair<T1, T2>> handleEvent(Event<? extends Object> input) throws Exception {
        if(uuid1.equals(input.getTag(TAG_KEY)))
            first = (Event<? extends T1>) input;
        else if(uuid2.equals(input.getTag(TAG_KEY)))
            second = (Event<? extends T2>) input;

        if(first == null || second == null)
            return null;
        // create result object
        Event<? extends Pair<T1,T2>> result = 
                input.createDerived(new Pair<T1,T2>(first.getData(), second.getData()));
        // copy tags over
        result = result.mergeTags(first.getTags());
        result = result.mergeTags(second.getTags());
        // ok, done
        return result.removeTags(TAG_KEY);
    }

    @Override
    public String toString() {
        return "MakePair(id1=" + uuid1 + ", id2=" + uuid2 + ")";
    }
}
