/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle, Sebastian Wrede, Jan Schaefer
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.xcf.fts.nodes.fusion;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import net.sf.xcf.fts.Event;
import net.sf.xcf.fts.annotations.Transform;
import net.sf.xcf.fts.Node;
import net.sf.xcf.fts.Pair;
import net.sf.xcf.fts.annotations.Fuse;

/**
 * A node that expects tagged input and fused data based on the tags.
 *
 * @author iluetkeb
 */
@Transform @Fuse
public class TaggedFusionNode implements Node<Object, Map<String, Object>> {

    /**
     * This is where data to be fused is collected.
     */
    protected final Map<String, Event<? extends Object>> fusedData = new HashMap<String, Event<? extends Object>>();
    /**
     * The tag that carries the source tagging information.
     */
    public static final String MAP_FUSE_NAME = "fusion.id";
    /**
     * Tag carrying association information. Will be copied to a list in the
     * output.
     */
    public static final String MAP_FUSE_ASSOCIATION_ID = "fusion.map.associator";

    protected Map<String, Object> getFusionMap() {
        final Map<String, Object> result = new HashMap<String, Object>(fusedData.size());
        for (Map.Entry<String, Event<? extends Object>> e : fusedData.entrySet()) {
            result.put(e.getKey(), e.getValue().getData());
        }
        return result;
    }

    protected Set<Object> getAssociationTag() {
        Set<Object> associations = new HashSet<Object>(fusedData.size());
        for (Event<? extends Object> evt : fusedData.values()) {
            Object associator = evt.getTag(MAP_FUSE_ASSOCIATION_ID);
            if (associator != null) {
                associations.add(associator);
            }
        }
        return associations;
    }

    @Override
    public Event<? extends Map<String, Object>> handleEvent(Event<? extends Object> input) {
        final String name = (String) input.getTag(MAP_FUSE_NAME);
        if (name == null) {
            throw new NullPointerException("Fusion ID must not be null, for " +
                    input);
        }
        fusedData.put(name, input);
        final Event<? extends Map<String, Object>> result = input.createDerived(getFusionMap()).
                removeTags(MAP_FUSE_NAME, MAP_FUSE_ASSOCIATION_ID);

        Set<Object> associators = getAssociationTag();
        if (associators.size() > 0) {
            return result.addTags(new Pair(MAP_FUSE_ASSOCIATION_ID, getAssociationTag()));
        } else {
            return result;
        }
    }
}
