/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle <ingo@fargonauten.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.xcf.fts.nodes.source;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import net.sf.xcf.fts.Event;
import net.sf.xcf.fts.EventFactory;
import net.sf.xcf.fts.Pair;
import net.sf.xcf.fts.TaggingEventFactory;
import net.sf.xcf.fts.Source;
import net.sf.xcf.fts.nodes.InjectorNode;

/**
 *
 * @author iluetkeb
 */
public class InjectorQueueHeadSource<T> extends AbstractSource<T>
        implements InjectorNode<T>, Source<T> {
    private final BlockingQueue<Event<? extends T>> queue;
    /** 
     * If receive last is set, every new item will clear the queue of 
     * previous ones.
     */
    private boolean receiveLast = false;
    /** 
     * Creates an InjectorQueueHeadSource that uses the given event factory
     * to create new Events from input objects.
     */
    public InjectorQueueHeadSource(EventFactory ef, int capacity) {
        super("InjectorQueueHeadSource", ef);
        queue = new LinkedBlockingQueue<Event<? extends T>>(capacity);
    }
    public InjectorQueueHeadSource(EventFactory ef) {
        super("InjectorQueueHeadSource", ef);
        queue = new LinkedBlockingQueue<Event<? extends T>>();
    }
    /** Creates InjectorQueueHead using a TaggingEventFactory.
     * @see net.sf.xcf.fts.TaggingEventFactory
     */
    public InjectorQueueHeadSource() {
        super("InjectorQueueHeadSource");
        queue = new LinkedBlockingQueue<Event<? extends T>>();
    }
    public InjectorQueueHeadSource(int capacity) {
        super("InjectorQueueHeadSource");
        queue = new LinkedBlockingQueue<Event<? extends T>>(capacity);
    }

    protected void setReceiveLast(boolean receiveLast) {
        this.receiveLast = receiveLast;
    }
    /**
     * 
     * Take the next event from the head of queue and return it, waiting for an
     * event to become available, if empty.
    */
    public Event<? extends T> next() throws Exception {
        return queue.take();
    }

    protected Event<? extends T> inject(Event<? extends T> evt) {
        if(receiveLast)
            queue.clear();
        
        if(!queue.add(evt))
            throw new IllegalStateException("Queue did not accept element " +
                    evt);
        return evt;
    }

    /**
     * Create new event from object and add it at the end of the queue.
     *
     * @throws IllegalStateException When the queue did not accept a new 
     *   element. This can not happen in this implementation.
     * @throws NullPointerException when data is null
     */
    public Event<? extends T> inject(T data) {
        return inject(createNewEvent(data));
    }

    public Event<? extends T> inject(T data, Pair<String, Object> ... tags) {
        return inject(createNewEvent(data).addTags(tags));
    }

}
