/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle <ingo@fargonauten.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.xcf.fts.nodes.source;

import java.util.concurrent.TimeUnit;
import net.sf.xcf.fts.Event;
import net.sf.xcf.fts.Source;
import net.sf.xcf.fts.annotations.NodeName;

/**
 * A source that queries its underlying source in regular intervals and sleeps
 * in the meantime.
 *
 * @author iluetkeb
 */
@NodeName("PeriodicSource")
public class PeriodicSource<T> implements Source<T> {
    private final Source<T> baseSource;
    private final long periodMillis;
    private long lastCall = System.currentTimeMillis();
    
    public PeriodicSource(Source<T> baseSource, long period, TimeUnit unit) {
        this.baseSource = baseSource;
        this.periodMillis = TimeUnit.MILLISECONDS.convert(period, unit);
    }
    public PeriodicSource(Source<T> baseSource, Long period) {
        this(baseSource, period, TimeUnit.MILLISECONDS);
    }
    public PeriodicSource(Source<T> baseSource, Long period, String unit) {
        this(baseSource, period, TimeUnit.valueOf(unit));
    }

    @Override
    public Event<? extends T> next() throws Exception {
        // compute the wait time as the configured period minus
        // the time that has already elapsed since the last call
        final long waitTime = periodMillis -(System.currentTimeMillis() - lastCall);
        if(waitTime > 0) {
            try {
                Thread.sleep(waitTime);
            } catch(InterruptedException ex) {
                return null;
            }
        }
        lastCall = System.currentTimeMillis();
        return baseSource.next();
    }

    @Override
    public Event<? extends T> handleEvent(Event<? extends Void> input) throws Exception {
        return next();
    }

    @Override
    public String toString() {
        return "PeriodicSource[" + periodMillis +
                "ms][" + baseSource + "]";
    }
}
