/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle, Sebastian Wrede, Jan Schaefer
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.xcf.fts.nodes.source;

import java.util.concurrent.atomic.AtomicInteger;
import net.sf.xcf.fts.annotations.NodeName;

/**
 * A source that generates the current time as a timestamp (in milisecons) each
 * time it is called. You can optionally specifiy a minimum interval between the
 * generated timestamps, in which case the source will block for some time if
 * necessary.
 * 
 * @author iluetkeb
 */
@NodeName("TimestampSource")
public class TimestampSource extends SourceAdapter<Long> {

    protected long lastTimestamp = System.currentTimeMillis();
    protected int interval;
    protected AtomicInteger cycles = new AtomicInteger();

    /**
     * Creates a timestamp source with no minimum interval.
     */
    public TimestampSource() {
        this(0);
    }

    /**
     * Creates a timestamp source with the given minimum interval.
     * 
     * @param interval
     *            The minimum interval, in ms.
     */
    public TimestampSource(int interval) {
        this(interval, -1);
    }

    /**
     * Creates a timestamp source with the given minimum interval and cycles.
     * 
     * TODO: What are the cycles? They aren't used for anything...
     * 
     * @param interval
     *            The minimum interval, in ms.
     * @param cycles
     *            ???
     */
    public TimestampSource(int interval, int cycles) {
        super("timestamp");
        this.interval = interval;
        this.cycles.set(cycles);
    }

    @Override
    public Long nextItem() throws Exception {
        final long diff = System.currentTimeMillis() - lastTimestamp;
        if (diff < interval) {
            Thread.sleep(interval - diff);
        }
        lastTimestamp = System.currentTimeMillis();
        cycles.getAndDecrement();
        // System.out.println("TimeStampSource " + System.nanoTime());
        return new Long(lastTimestamp);
    }
}
