/*
 * Copyright 2008, 2009 Bielefeld University
 * Copyright 2008, 2009 Ingo Luetkebohle <ingo@fargonauten.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.xcf.fts.nodes.transform.xml;

import java.util.List;
import java.util.Map;
import net.sf.xcf.fts.annotations.NodeName;
import net.sf.xcf.fts.nodes.transform.AbstractTransformNode;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Node;
import nu.xom.Text;
import org.jaxen.JaxenException;
import org.jaxen.XPath;

/**
 * Looks for a comparison xpath in a target document and a list of background 
 * documents. If a match is found, a second xpath specifies what to copy from
 * the background document and a third xpath where to put it.
 *
 * If no match is found, the input document is returned unchanged.
 *
 * The target document is modified in-place.
 *
 * @author iluetkeb
 */
@NodeName("CompareAndFuse")
public class CompareAndFuseTransform extends AbstractTransformNode<Map<String, Object>, Document> {
    private final XPath comparison;
    private final XPath copySelector;
    private final XPath insertionSelector;

    /**
     * Create a background fusion node.
     *
     * @param comparison Selects a value that must be equal between documents
     * @param copySelector Selects nodes to copy into targte
     * @param insertionSelector Select an Element (required!) to append copy data to.
     */
    public CompareAndFuseTransform(XPath comparison, XPath copySelector,
            XPath insertionSelector) {
        this.comparison = comparison;
        this.copySelector = copySelector;
        this.insertionSelector = insertionSelector;
    }

    /**
     * Key for the region to add info to. Must be of type Document.
     */
    public static final String KEY_TARGET = "fuse.target";
    /** Key for the list of background documents. Must be a
     * List containing Documents. */
    public static final String KEY_BACKGROUND = "fuse.background";

    protected String getComparisonValue(Document target) throws JaxenException {
        Node value = (Node)comparison.selectSingleNode(target);
        if(value != null)
            return value.getValue();
        else
            return null;
    }

    protected Document fuse(Document target, Document bgDoc) throws JaxenException {
        target = (Document) target.copy();
        Element insertPoint = (Element) insertionSelector.selectSingleNode(target);
        List copyValues = copySelector.selectNodes(bgDoc);
        for(Object o : copyValues) {
            if(o instanceof Node)
                insertPoint.appendChild(((Node)o).copy());
            else
                insertPoint.appendChild(new Text(o.toString()));
        }
        return target;
    }
    @Override
    protected Document transform(Map<String, Object> input) throws Exception {
        Document target = (Document) input.get(KEY_TARGET);
        List<Document> background = (List<Document>) input.get(KEY_BACKGROUND);
        if(background == null)
            return target;

        final String comparator = getComparisonValue(target);
        if(comparator == null)
            return target;
        for(Document bgDoc : background) {
            final String bgValue = getComparisonValue(bgDoc);
            if(comparator.equals(bgValue)) {
                return fuse(target, bgDoc);
            }
        }

        return target;
    }
}
